;--------------------------------------------------------------------
;
;  e3.asm v1.61  Copyright (C) 2000,2001 Albrecht Kleine <kleine@ak.sax.de>
;
;  This program is free software; you can redistribute it and/or modify
;  it under the terms of the GNU General Public License as published by
;  the Free Software Foundation; either version 2 of the License, or
;  (at your option) any later version.
;
;  This program is distributed in the hope that it will be useful,
;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;  GNU General Public License for more details.
;
;  You should have received a copy of the GNU General Public License
;  along with this program; if not, write to the Free Software
;  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;
;----------------------------------------------------------------------
%define WS 1
%define EM 2
%define PI 4
%define VI 8
%define NE 16
%define DEFAULT_MODE WS			;<---- select one of WS, EM, NE, PI, VI
%define MAKE_BACKUP
%define LESSWRITEOPS
%define USE_MATH
;
%define BEEP_IN_VI
;
;
;-------
;
%ifdef EX
 %define USE_EX_NO_SED
 %define EX_PATH '/usr/bin/ex'		;(ex is usually a symlink to vi)   [ old was /bin/ex ]
%else
 %undef USE_EX_NO_SED
 %define SEDPATH '/bin/sed'		;DEFAULT because we want to be INDEPENDENT of vi-derivates
%endif

;-------
%ifdef LINUX
 %define CURSORMGNT			;switch cursor depending of 'INSERT'-mode
%else
 %undef CRIPLED_ELF
%endif
;
;----------------------------------------------------------------------
section .text				;here it goes....
bits 32

%ifdef CRIPLED_ELF
;
; building e3 via "nasm -f bin ...."  using an idea from
;"A Whirlwind Tutorial on Creating Really Teensy ELF Executables for Linux"
;
	org 0x08048000
ehdr:	db 0x7F, "ELF", 1, 1, 1, 0	;Elf32_Ehdr starts here
	dd 0,0
	dw 2				;e_type
	dw 3				;e_machine
	dd 1				;e_version
	dd _start			;e_entry
	dd phdr1- $$			;e_phoff
	dd 0				;e_shoff
	dd 0				;e_flags
	dw ehdrsize			;e_ehsize
	dw phdrsize			;e_phentsize
	dw 2				;e_phnum
phdr1:					;Elf32_Phdr starts here
	dd 1				;both p_type and e_shentsize,e_shnum
	dw 0				;both p_offset and e_shstrndx
ehdrsize equ $ - ehdr
	dw 0
	dd $$				;p_vaddr
	dd $$				;p_paddr
	dd filesize			;p_filesz
	dd filesize			;p_memsz
	dd 5				;p_flags i.e. READ/EXECUTE
	dd 0;  0x1000			;p_align
phdrsize equ $ - phdr1
phdr2:					;another Elf32_Phdr starts here
	dd 1				;p_type
	dd filesize
	dd $$+filesize
	dd $$+filesize
	dd 0				;p_filesz
	dd bsssize			;p_memsz
	dd 6				;p_flags i.e. READ/WRITE
	dd 0;  0x1000			;p_align
%endif
;----------------------------------------------------------------------
;
%ifdef LESSWRITEOPS			;constraints
	%define LESSWRITEOPS_OR_CURSORMGNT
%endif
%ifdef CURSORMGNT
	%define LESSWRITEOPS_OR_CURSORMGNT
%endif

stdtxtlen	equ 10			;do not move this to EOF: code size would increase
%ifdef FREEBSD
O_WRONLY_CREAT_TRUNC equ 601h		;see fcntl.h
%else
O_WRONLY_CREAT_TRUNC equ 1101q
%endif
O_RDONLY	equ 0
PERMS		equ 644q
stdin		equ 0
stdout 		equ 1
optslen		equ 127
TAB		equ 8
TABCHAR		equ 09h
SPACECHAR	equ ' '
CHANGED		equ '*'
UNCHANGED	equ SPACECHAR
LINEFEED	equ 0ah
NEWLINE		equ LINEFEED
RETURN		equ 0dh
SEDBLOCK	equ 4096
;
;----------------------------------------------------------------------
global _start
;
; start with OUTER editor loop
;
_start:	call SetTermStruc
%ifdef BEOS
	pop eax
	pop ebx				;args counter (1....x)
	pop esi
	mov esi,[esi]			;points to our_name0[args]0env.....
	cld
prog:	lodsb
	or al,al
	jne prog
	mov edx,esi			;store arg ptr
	lea esi,[esi-5]
	mov eax,mode
	mov byte [eax],DEFAULT_MODE	;store default editor mode
	call SetEditMode
	xor esi,esi			;init for 'no args'
	dec ebx
	jz noarg
	mov esi,edx
noarg:
%else
	pop edx				;Linux: arguments #
;-------
	pop esi				;Linux: argv[0]
	cld
prog:	lodsb
	or al,al
	jne prog			;get program name
	lea esi,[esi-5]
	mov eax,mode
	mov byte [eax],DEFAULT_MODE	;store current editor mode
	call SetEditMode
;-------
	pop esi				;Linux: esi points to first arg (filename)
%endif
;-------
%ifdef CURSORMGNT
	or esi,esi
	jz moreenv
morearg:pop ecx				;arguments until NULL
	or ecx,ecx
        jnz morearg
;-------
moreenv:pop ecx
	jecxz ReStart
	cmp dword[ecx],'TERM'		;a short test for "TERM=linux"
	jnz moreenv
	cmp dword[ecx+5],'linu'
	jnz ReStart
	add byte[revvoff],boldlen	;special inverse cursor on linux terminals
%endif
;-------
ReStart:call NewFile
	jc E3exit
MainCharLoop:call ChkCursPos
	call DispNewScreen
	call RestoreStatusLine
	call HandleChar
	cmp byte [endeedit],0
	je MainCharLoop
	xor esi,esi			;just like if no arg is present
	cmp byte [endeedit],2
	je ReStart			;^KD repeat edit using another file
E3exit:	call KursorStatusLine
	mov ecx,text			;enter next line on terminal NEWLINE is @ byte [text]
	xor edx,edx
	inc edx				;mov edx,1
	call WriteFile0
;-------
%ifdef BEOS
	mov ecx,8001h			;TCSETA
%else
%ifdef LINUX
	mov ecx,5402h			;TCSETS asm/ioctls.h
%else
	mov ecx,802c7414h		;TIOCSETA
%endif
%endif
	call IOctlTerminal0		;restore termios settings
	jmp Exit
;----------------------------------------------------------------------
;
; MAIN function for processing keys
;
HandleChar:call ReadChar
	cmp ah,0xFF			;normal chars get 0xFF in ah
	jne near ExtAscii		;go handling Cursor-Keys
	mov esi,mode
	test byte [esi], EM | PI
	jz NO_EM01
	cmp al,11
	je IsCtrlK
	mov byte [EmaCtrlK],0
IsCtrlK:cmp al,13h			;^S
	je IsCtrlS
	cmp al,12h			;^R
	je IsCtrlS
	mov byte [EmaCtrlS],0
IsCtrlS:
NO_EM01:cmp byte [esi],VI
	jz ISVI1
	cmp al,32			;in WS,EM,PI,NE: handle control chars
	jae NormChar
	mov bl,al
	add bl,jumps1
	cmp byte [esi],WS
	je CJump
	add bl,32
	cmp byte [esi],EM
	je CJump
	add bl,32
	cmp byte [esi],PI
	je CJump
	add bl,32
CJump:	jmp CompJump2
ISVI1:					;in VI: most control is done in command mode...
	cmp al,7			;... so maintaining another table for <Return>...
	je near KeyDel			;... <Del> and <DelLeft> is useless
	cmp al,8
	je near KeyDell
	cmp al,RETURN
	je near KeyRet
;-------
NormChar:test byte [mode], EM | PI
	jz NOEM0
	mov byte [showblock],0
NOEM0:
	call CheckMode
	jnz OverWriteChar
NormCh2:push eax
	xor eax,eax
	inc eax
	call InsertByte
	pop eax
	jc InsWriteEnd			;error: text buffer full
OverWriteChar:cld
	stosb
OverW2:	mov byte [changed],CHANGED
InsWriteEnd:ret
;-------
KeyVICmdr:call ReadOneChar		;repl one char (except newline)
	cmp byte [edi],NEWLINE
	je InsWriteEnd
	cmp al,RETURN
	jnz KeyVICmdr1
	mov al,NEWLINE
KeyVICmdr1:mov byte [edi],al
	jmp short OverW2
;-------
KeyEmaCtrlQ:mov esi,asknumber
	call GetOctalToInteger
	jbe InsWriteEnd
	xchg eax,ecx			;using decimal input for ASCII value
	cmp eax,256
	jb NormCh2
	ret
;-------
;
; helper for HandleChar
;
CtrlKMenu:mov ebx,Ktable
	mov al,'K'
	jmp short Menu
CtrlQMenu:mov ebx,Qtable
	jmp short PicoQM
PicoJMenu:mov ebx,PicoJtable
	mov al,'J'
	jmp short Menu
PicoQMenu:mov ebx,PicoQtable
PicoQM:	mov al,'Q'
	jmp short Menu
CtrlXMenu:mov ebx,Xtable
	mov al,'X'
Menu:	mov ecx,2020205eh
	mov ch,al
	call MakeScanCode
	jnc InsWriteEnd			;if no valid scancode
ExtAscii:mov bl,ah			;don't use al (carries char e.g. TAB)
	cmp bl,jumps1
	jae InsWriteEnd
	xor eax,eax
	mov byte [EmaCtrlK],al
	mov byte [EmaCtrlS],al
CompJump2:mov bh,0
	lea ebx,[bx]			;1 byte shorter than 'and ebx,0ffh'
	movzx ebx,word [2*ebx+jumptab1] ;2*ebx is due 2 byte per entry
	add ebx,_start			;offset inside code
;-------
	call ebx			;the general code jump dispatcher
;-------
	cmp byte [numeriere],1		;after return from functions...
	jnz BZNret			;...decide whether count current line number
	push edi
	mov esi,sot
	xchg esi,edi
	xor edx,edx
BZNLoop:inc edx				;edx=linenr counter
	call LookForward
	inc edi				;point to start of next line
	cmp edi,esi
	jbe BZNLoop
	mov [linenr],edx
	pop edi
	mov byte [numeriere],0
BZNret:	ret
;-------
MakeScanCode:call WriteTwo		;ebx expects xlat-table
	push ebx
	call GetChar
	pop ebx
	and al,01fh
	cmp al,27
	jnb EndeRet
	xlatb
	mov ah,al			;returns pseudo "scancode"
EndeRet:ret				;exception: ok=cy here
;----------------------------------------------------------------------
;
; processing special keys: cursor, ins, del
;
KeyRetNoInd:xor eax,eax
	jmp short KeyRetNInd
KeyRet:	call CheckMode
	jnz  OvrRet
	call CountToLineBegin		;set esi / returns eax
	inc esi
	inc esi
	or eax,eax
	jz KeyRetNInd
	xor eax,eax
	dec eax
KeyRetSrch:inc eax			;search non (SPACE or TABCHAR)
	cmp byte [esi+eax],SPACECHAR
	je KeyRetSrch
	cmp byte [esi+eax],TABCHAR
	je KeyRetSrch
KeyRetNInd:push esi
	push eax			;eax is 0 or =indented chars
	call GoDown
	pop eax
	push eax
	inc eax				;1 extra for 0ah
	call InsertByte
	pop ecx				;# blanks
	pop esi				;where to copy
	jc SimpleRet
	inc dword [linenr]
	mov al,NEWLINE
	cld
	stosb
	jecxz SimpleRet
	rep movsb			;copy upper line i.e. SPACES,TABS into next
SimpleRet:ret
OvrRet:	xor eax,eax
	mov [ch2linebeg],eax
	jmp short DownRet
;-------
KeyDown:call CountColToLineBeginVis
DownRet:call GoDown
	call LookLineDown
	jmp short SetColumn
;-------
KeyUp:	call GoUp
	call CountColToLineBeginVis
	call LookLineUp
	jmp short SetColumn
;-------
KeyHalfPgUp:call CountColToLineBeginVis
	call LookHalfPgUp
	jmp short SetColumn
;-------
KeyHalfPgDn:call CountColToLineBeginVis
	call LookHalfPgDn
	jmp short SetColumn
;-------
KeyScrollUp:call CountColToLineBeginVis
	call LookScrUp
	jmp SetColumn
KeyScrollDn:call CountColToLineBeginVis
	call LookScrDn
	jmp SetColumn
;-------
KeyPgUp:call CountColToLineBeginVis
	call LookPageUp
	jmp short SetColumn
;-------
KeyPgDn:call CountColToLineBeginVis
	call LookPgDown			;1st char last line
;-------
SetColumn:mov ecx,[ch2linebeg]		;maximal columns
	xor edx,edx			;counts visible columns i.e. expand TABs
	dec edi
SCloop:	inc edi
	cmp edx,ecx			;from CountColToLineBeginVis
	jae SCret
	cmp byte [edi],NEWLINE		;don't go beyond line earlier line end
	jz SCret
	cmp byte [edi],TABCHAR
	jz SCtab
	inc edx				;count columns
	jmp short SCloop
SCtab:	call SpacesForTab
	add dl,ah
	cmp edx,ecx			;this tab to far away right?
	jna SCloop			;no
SCret:	ret
;-------
KeyHome:call CountToLineBegin
	sub edi,eax
	ret
;-------
KeyEnd:	call CountToLineEnd
	add edi,eax			;points to a 0ah char
	ret
;-------
KeyIns:	not byte [insstat]
	xor eax,eax
	cmp byte [mode],VI
	jnz KeyIns2
	inc eax
	cmp byte [VICmdMode],al
	jne KeyIns2
	mov byte [insstat],al
	call KeyVImode0
KeyIns2:cmp byte[mode],EM
	jnz KeyIns3
	mov byte [showblock],al
KeyIns3:ret
;-------
KeyVICmdJ:call KeyEnd
	jmp short KeyDel
;-------
KeyDell:call KeyLeft
	jz KeyDell2
KeyDel:	cmp edi,ebp
	jnb KeyLeftEnd
	xor eax,eax			;delete one @ cursor
	inc eax
	jmp DeleteByte
KeyDell2:cmp edi, sot			;delete newline char
	jbe KeyLeftEnd
	dec dword [linenr]
	dec edi
	jmp KeyCtrlT1
;-------
KeyLeft:cmp byte [edi-1],NEWLINE
	jnz KeyLNoMargin
	cmp edi,sot			;i.e. CheckBof
	je KeyLeftEnd
	cmp byte [mode],VI
	je KeyLeftEnd			;no more line wrap around in vi mode
	call GoUp
KeyLNoMargin:dec edi
KeyLeftEnd:ret
;-------
KeyEmaCtrlT:cmp edi,sot			;xchg 2 chars
	jbe KeyRightEnd
	cmp byte [edi],NEWLINE
	jnz KECT
	dec edi
KECT:	mov al, byte [edi]
	xchg al, byte [edi-1]
	mov byte [edi],al
	mov byte [changed],CHANGED
;-------
KeyRight:cmp byte [edi],NEWLINE
	jnz KeyRNoMargin
	call CheckEof
	jae KeyRightEnd
	cmp byte [mode],VI
	je KeyRightEnd			;no more line wrap around in vi mode
	call GoDown
KeyRNoMargin:inc edi
KeyRightEnd:ret
;-------
KeyCLeft3:cmp edi, sot
	jbe KeyCLEnd
	dec edi
KeyCtrlQW:cmp byte [edi-1],NEWLINE
	jz KeyCLeft3
	dec edi
	cmp byte [edi],2fh
	jbe KeyCtrlQW
	cmp byte [edi-1],2fh
	ja KeyCtrlQW
KeyCLEnd:ret
;-------
KeyCRight3:call CheckEof
	jae KeyCREnd
	inc edi
KeyCtrlQZ:cmp byte [edi],NEWLINE	;was KeyCtrlRight
	jz KeyCRight3
	inc edi
	cmp byte[mode],EM
	jz ISEM2
	cmp byte [edi],2fh		;ws stops at word begin
	jbe KeyCtrlQZ
	cmp byte [edi-1],2fh
	jmp short ISEM22
ISEM2:	cmp byte [edi-1],2fh		;em stops after end
	jbe KeyCtrlQZ
	cmp byte [edi],2fh
ISEM22:	ja KeyCtrlQZ
KeyCREnd:ret
;-------
KeyVIcmde3:call CheckEof		;end of word (vi only)
	jae KeyCREnd
	inc edi
KeyVIcmde:cmp byte [edi],NEWLINE
	jz KeyVIcmde3
	inc edi
	cmp byte [edi],2fh
	jbe KeyVIcmde
	cmp byte [edi+1],2fh
	ja KeyVIcmde
	ret
;-------
KeyEmaCtrlO:xor eax,eax
	inc eax
	call InsertByte
	jc KeyRightEnd
	mov al,NEWLINE
	mov byte [edi],al
SimpleRet9:ret
;----------------------------------------------------------------------
;
; processing special keys from the WS's Ctrl-Q menu
;
KeyCtrlQE:call LookPgBegin		;goto top left on screen
	call KursorFirstLine
	jmp short KCtKV1
;-------
KeyCtrlQX:call LookPgEnd		;1st goto last line on screen
	call KeyEnd			;2nd goto line end
	call KursorLastLine
	jmp short KCtKV1
;-------
KeyCtrlQV:cmp byte [bereitsges],0	;goto last ^QA,^QF pos
	jz KeyCREnd
	mov edi,[oldQFpos]
KCtKV1:	jmp CQFNum
;-------
KeyEmaAltPer:push byte 1		;s&repl
	pop dword[vorwarts]
	mov byte[grossklein],0dfh
;-------
KeyCtrlQA:mov byte [bereitsges],2
	call AskForReplace
	jc SimpleRet9
CQACtrlL:push edi
	call FindText
	jc near CtrlQFNotFound
	mov eax,[suchlaenge]
	call DeleteByte
	mov eax,[repllaenge]
	call InsertByte
	mov esi,replacetext
	call MoveBlock
	jmp near  CQFFound
;-------
KeyVIbsearch:mov byte[grossklein],0ffh
	jmp short KECR
KeyEmaCtrlR:mov byte[grossklein],0dfh
KECR:	push byte -1
	jmp short KeyECtS1
KeyVIfsearch:mov byte[grossklein],0ffh
	jmp short KECS
;-------
KeyPiCtrlJT:cmp byte [bereitsges],2
	jz CQACtrlL
	ret
;-------
KeyEmaCtrlS:mov byte[grossklein],0dfh
KECS:	push byte 1
KeyECtS1:pop dword[vorwarts]
	mov [EmaMark],edi
	mov byte [showblock],0
;------- cont
KeyCtrlQF:
	cmp byte [mode],EM
	jnz NO_EM04
	cmp byte [EmaCtrlS],1
	jz KeyCtrlL
NO_EM04:push dword [suchtext]		;store begin of old find_text
	mov byte [bereitsges],1
	call AskForFind
	pop ebx
	pushf
	test byte [mode], VI | PI
	jz NO_VIPI01
	popf
	or eax,eax			;jmp if user entered a new find_text
	jnz QFpico
	or bl,bl			;jmp if no old find text available
	jz CtrlQFEnd
	mov byte [suchtext],bl		;restore last find_text
QFpico:	mov [PicoSearch],edi		;store begin of search (because wrap around EOF)
	jmp short CQFCtrlL
;-------

NO_VIPI01:popf
	jc CtrlQFEnd
CQFCtrlL:push edi
	call FindText
	mov byte [EmaCtrlS],1
	jc CtrlQFNotFound
CQFFound:mov [oldQFpos],edi
	pop esi				;dummy
CQFNum:	mov byte [numeriere],1
	ret
CtrlQFNotFound:pop edi
CtrlQFEnd:ret
;-------
KeyCtrlL:mov eax,[bereitsges]		;2^QA   1^QF   0else
	dec eax
	jz CQFCtrlL
	test byte[mode],WS | NE | PI
	jz SimpleRet4
	dec eax
	jz near CQACtrlL
SimpleRet4:ret
;-------
KeyCtrlQC:mov edi,ebp
	jmp short CQFNum
;-------
KeyVIcmd1:call ReadOneChar
	cmp al,'G'
	je KeyCtrlQR
	ret
ViSpecial:jecxz KeyCtrlQR
	jmp short KCQI
KeyNedCtrlA:mov [EmaMark],ebp
	mov byte [showblock],1
KeyCtrlQR:mov edi,sot
	jmp short CQFNum
;-------
KeyCtrlQP:mov ecx,[veryold]
	cmp ecx,ebp
	ja SimpleRet4
	mov edi,ecx
JmpCQFN3:jmp short CQFNum
;-------
KeyCtrlQB:xchg eax,edi
	mov edi,[blockbegin]
CtrlQB2:or edi,edi			;exit if no marker set
	jnz CQFNum
	xchg edi,eax
	ret
;-------
KeyCtrlQK:xchg eax,edi
	mov edi,[blockende]
	jmp short CtrlQB2
;-------
KeyCtrlQI:mov esi,asklineno
	call GetAsciiToInteger
	jbe CtrlQFEnd			;CY or ZR set
KCQI:	mov edi,sot
	call LookPD2
JmpCQFN:jmp short JmpCQFN3
;-------
KeyCtrlQDel:call KeyLeft		;delete all left of cursor
	call CountToLineBegin
	sub edi,eax
	call DeleteByteCheckMarker
	jmp short KeyCtrlT1
;-------
KeyVICmdD:mov byte [VInolinebased],1
KeyCtrlQY:call CountToLineEnd
	jmp short CtrlTEnd1
;-------
KeyCtrlY:call CountToLineBegin
	sub edi,eax        		;edi at begin
	call CountToLineEnd
	cmp byte[mode],WS
	jnz NO_WS01
	call DeleteByteCheckMarker
	jmp short KeyCtrlT1
NO_WS01:test byte [mode], VI | PI
	jz KeyCtrlT
	lea ecx,[edi+eax]
	cmp ecx,ebp
	jz CtrlTEnd1			;do not delete pending 0Ah (Sun Nov 12)
	inc eax
	jmp short CtrlTEnd1
;-------
KeyCtrlT:call CountToWordBegin
	cmp byte[mode],EM
	jnz NO_EM05
KeyEmaCtrlK:call CountToLineEnd
NO_EM05:cmp byte [edi],NEWLINE
	jnz CtrlTEnd1
KeyCtrlT1:xor eax,eax
	inc eax				;1 for 0ah
CtrlTEnd1:call CheckEof
	jz SimpleRet3
	cmp byte[mode],WS
	jz near DeleteByteCheckMarker
	mov esi,edi
	call KeyEmaAltW2
	jmp DeleteByte
;----------------------------------------------------------------------
;
; processing special Keys from WS's Ctrl-K menu
;
KeyCtrlKY:call CheckBlock
	jc SimpleRet3			;no block: no action
	mov eax,[blockende]
	mov edi,esi			;esi is blockbegin (side effect in CheckBlock)
	sub eax,esi			;block length
	call DeleteByte			;out ecx:=0
	mov [blockende],ecx		;i.e. NO block valid now
	mov [blockbegin],ecx
JmpCQFN2:jmp JmpCQFN
;-------
KeyCtrlKH:xor byte [showblock],1 	;flip flop
SimpleRet3:ret
KeyCtrlKK:mov [blockende],edi
	jmp short KCKB
;-------
KeyCtrlKW:call CheckBlock
	jc SimpleRet2   		;no action
	call SaveBlock
	jmp CtrlKREnd
;-------
KeyCtrlKC:call CopyBlock
	jc SimpleRet2
CtrlKC2:mov [blockbegin],edi
	add eax,edi
	mov [blockende],eax
	ret
;-------
KeyCtrlKV:call CopyBlock
	jc SimpleRet2
	push edi
	cmp edi,[blockbegin]
	pushf
	mov edi,[blockbegin]
	call DeleteByte
	neg eax				;(for optimizing eax is negated there)
	popf
	pop edi
	jb CtrlKC2
	mov [blockende],edi
	sub edi,eax
KeyCtrlKB:mov [blockbegin],edi
KCKB:	mov byte [showblock],1
SimpleRet2:ret
;-------
KeyCtrlXX:mov ecx,[EmaMark]
	jecxz SimpleRet3
	call KeyEmaMark
	mov edi,ecx
	call KeyEmaCtrlL
	jmp JmpCQFN
;-------
KeyVICmdm:call ReadOneChar
	cmp al,'a'
	jne SimpleRet2
	push edi
	call KeyHome
	call KeyEmaMark
	pop edi
	ret
KeyEmaMark:mov byte [showblock],1
	mov [EmaMark],edi
	ret
;-------
KeyCtrlKR:call ReadBlock
	jc CtrlKREnd
	call KeyCtrlKB
	add ecx,edi
	mov [blockende],ecx
	test byte [mode],EM | NE
	jz NO_EM03
	mov [EmaMark],ecx
	mov byte [showblock],0
NO_EM03:cmp byte [mode],PI
	jnz CtrlKREnd
	mov edi,ecx			;in PI: cursor at end of read file
CtrlKREnd:jmp RestKursPos
;-------
KeyEmaCtrlXF:cmp byte [changed],UNCHANGED
	jz KECF
	mov esi,asksave2
	call DE1
	call RestKursPos
	call CheckUserAbort
	jz CtrlKSEnd
	and al,0dfh
	cmp al,'N'			;N for request NOT SAVE changes
KECF:	jz KCKD2
	jmp short KeyCtrlKD
;-------
KeyEmaCtrlXW:call GetBlockName
	jc near CtrlKSEnd
	mov esi,blockpath
XW1:	cld
	pusha
	mov edi,filepath
XW0:	lodsb
	stosb				;copy to blockpath to filepath
	or al,al
	jne XW0
	stosb
	popa
KeyCtrlKS0:mov byte [changed],CHANGED	;to save it really
;-------
KeyCtrlKS:call SaveFile
	pushf				;(called by ^kd)
	call RestKursPos
	popf
	jc CtrlKSEnd
	mov byte [changed],UNCHANGED
CtrlKSEnd:ret
;-------
KeyCtrlKD:call KeyCtrlKS
	jc KeyKXend
KCKD2:	mov byte [endeedit],2
	ret
;-------
KeyCtrlKQ:cmp byte [changed],UNCHANGED
	jz KCKXend
	mov esi,asksave
	call DE1
	call RestKursPos
	call CheckUserAbort
	jz CtrlKSEnd
	and al,0dfh
	cmp al,'N'			;N for request NOT SAVE changes
	jz KCKXend
	cmp al,'L'			;L for SAVE and LOAD a new file
	jnz KeyCtrlKX
	call KCKXend
KeyCtrlKX:call KeyCtrlKS
	jc CtrlKSEnd
KCKXend:inc byte [endeedit]
KeyKXend:ret
;----------------------------------------------------------------------
;
; some minimal limited vi specials in command mode
;
KeyVICmdW:lea esi,[ecx+2]
	cmp byte [esi],SPACECHAR
	ja XW1
	ret
;-------
VINoLineCmd:mov eax,[ecx]
	cmp ax,'$'			;No line number, but EOF
	je near KeyCtrlQC
	cmp ax,'w'			;save
	je near KeyCtrlKS
	cmp ax,'w!'			;save
	je near KeyCtrlKS0
	cmp ax,'wq'
KVI_KX:	je KeyCtrlKX
	cmp ax,'w '			;save as ... and continue
	je KeyVICmdW
	cmp ax,'q'
	je KeyCtrlKQ
	cmp ax,'q!'
	je KCKXend
	cmp ax,'e '			;edit another ..
	je near KeyVICmdE
	cmp ax,'h'
	je near KeyHelp
%ifndef LINUX
	ret
%else
	jmp KeyVICmdtemp
%endif
;-------
KeyVI1Char:call KeyHome
	cmp byte [edi],SPACECHAR
	ja KFC2
KFC1:	cmp byte [edi],NEWLINE
	jz KFC2
	inc edi
	cmp byte [edi],SPACECHAR
	jbe KFC1
	cmp byte [edi-1],SPACECHAR
	ja KFC1
KFC2:	ret
;-------
KeyVICmdS:call KeyHome
	call KeyEmaCtrlK		;not quite ok in 'P'/'p' commands
	mov byte [VInolinebased],1
	jmp short KeyVICmdI
KeyVICmdd:call ReadOneChar
	cmp al,'d'			;"delete"
	mov byte [VInolinebased],0
	je near KeyCtrlY
	cmp al,"'"			;"by line"
	jne KFC2
	call ReadOneChar
	cmp al,'a'			;d'a
	jne KFC2
	mov ecx,[EmaMark]		;don't go forther if no mark set
	jecxz KFC2
	mov ecx,[showblock]
	jecxz KFC2
	call KeyHome
	call KeyDown
	call KeyEmaCtrlW
	jmp short JmpCQFn
;-------
KeyVICmdZ:call ReadOneChar
	cmp al,'Z'
	je near KVI_KX
	ret
KeyVICmdz:call ReadOneChar
	cmp al,'.'
	je near KeyEmaCtrlL
KVCz:	ret
;-------
KeyVICmdI:call KeyVI1Char
	jmp short KeyVImode0
;-------
KeyVICmdp:mov ecx,[EmaKiSize]		;check this before call KeyEmaCtrlY
	jecxz KVCz
	cmp byte [VInolinebased],1
	jz KeyVICmdpnLB
	call OvrRet			;ugly
KeyVICmdP:mov ecx,[EmaKiSize]		;check this before call KeyEmaCtrlY
	jecxz KVCz
	cmp byte [VInolinebased],1
	jz KeyVICmdPnLB
	call KeyHome
KeyVICP2:push edi
	call KeyEmaCtrlY
	pop edi
JmpCQFn:jmp CQFNum
;-------
KeyVICmdR:mov byte [insstat],254	;i.e "not 1"
	jmp short KeyVImode0
KeyVICmdO:call KeyHome
	call KeyRet
	call KeyUp
	jmp short KeyVImode0
KeyVICmdo:call KeyEnd
	call KeyRet
	jmp short KeyVImode0
KeyVICmdA:call KeyEnd
	jmp short KeyVImode0
KeyVIcmda:call KeyRight
KeyVIcmdi:mov byte [insstat],1
KeyVImode0:push byte 0
	jmp short KVim1
KeyVICmdC:call KeyEmaCtrlK
	mov byte [VInolinebased],1
	jmp short KeyVImode0

;-------
KeyVICmdpnLB:call KeyVICP2
	jmp KeyEnd
KeyVICmdPnLB:call KeyLeft		;not_Line_Based mode
	call KeyVICmdpnLB
	jmp KeyLeft
;-------
KeyVIex:mov esi,extext
	call InputStringWithMessage
	pushf
	call RestKursPos
	popf
	jc Kviex
	mov esi, optbuffer
	xor edx,edx
	mov ecx,eax
	jecxz Kviex
;-------
	cld
CheckDig:lodsb				;check for line number entered
	cmp al,'0'
	adc edx,0
	cmp al,':'
	cmc
	adc edx,0
	loop CheckDig
	mov ecx, optbuffer
	or edx,edx
	jnz near VINoLineCmd
	call GetAsciiToInteger
	jmp ViSpecial			;due short jumps there
;-------
KeyVImode1:push byte 1
KVim1:	pop eax
	mov byte [VICmdMode],al
Kviex:	ret
;---------------------------------------------------------------------
;
; the general PAGE DISPLAY function: called after any pressed key
;
; side effect: sets 'columne' for RestoreStatusLine function (displays column)
; variable kurspos: for placing the cursor at new position
; register bh counts lines
; register bl counts columns visible on screen (w/o left scrolled)
; register edx counts columns in text lines
; register ecx screen line counter and helper for rep stos
; register esi text index
; register edi screen line buffer index
;
DispNewScreen:test byte [mode], EM | PI | NE
	jz NoEmBlock
	mov ecx,[showblock]		;transfering Emacs's mark/point into....
	jecxz NoEmBlock			;....WS's block display system
	mov ecx,[EmaMark]
	jecxz NoEmBlock
	mov eax,edi
	cmp ecx,eax
	jb EmBlock
	xchg eax,ecx
EmBlock:mov [blockbegin],ecx
	mov [blockende ],eax
;-------
NoEmBlock:call GetEditScreenSize	;check changed tty size
	xor eax,eax
	mov byte[isbold],al
	mov byte[inverse],al
	mov [zloffst],eax
	mov [columne],eax
	mov [fileptr],edi		;for seeking current cursor pos
	call CountColToLineBeginVis	;i.e. expanding TABs
	mov ebx,[columns]
	lea ebx,[ebx-4]			;03 Jun 2001
	cmp eax,ebx
	jb DispShortLine
	sub eax,ebx
	;;;shr eax,4	;/16
	inc eax
	;;;shl eax,4	;*16
	mov [zloffst],eax
DispShortLine:call LookPgBegin 		;go on 1st char upper left on screen
	mov esi,edi			;esi for reading chars from text
	mov ecx,[lines]
	jecxz Kviex
	cld
	mov bh,-1			;first line
DispNewLine:inc bh			;new line
	mov edi,screenline		;line display buffer
	xor edx,edx			;reset char counter
	mov bl,0 			;reset screen column to 0
%ifdef LESSWRITEOPS
	call SetColor2			;set initial character color per each line
%endif
DispCharLoop:
	cmp esi,[fileptr]		;display char @ cursor postion ?
	jnz DispCharL1
	cmp byte[tabcnt],0
	jnz DispCharL1
	mov [kurspos],ebx
	mov byte [columne],bl
	mov eax,[zloffst]		;chars scrolled left hidden
	add [columne],eax
%ifdef CURSORMGNT
	stc
	call SetInverseStatus
	jnc DispEndLine
%endif
DispCharL1:call SetColor		;set color if neccessary
;-------
DispEndLine:cmp esi,ebp
	ja FillLine			;we have passed EOF, so now fill rest of screen
	cmp byte[tabcnt],0
	jz ELZ
	dec byte[tabcnt]
	jmp short ELZ2
ELZ:	cmp esi,ebp
	jnz ELZ6
	inc esi				;set esi>ebp will later trigger  "ja FillLine"
	jmp short ELZ2
ELZ6:	lodsb
	cmp al,TABCHAR
	jnz ELZ3
	call SpacesForTab		;ah = space_up_to_next_tab location
	dec ah				;count out the tab char itself
	mov byte[tabcnt],ah
ELZ2:	mov al,SPACECHAR
ELZ3:	cmp al,NEWLINE
	jz FillLine
	cmp al,SPACECHAR
	jae ELZ9			;simply ignore chars like carriage_return etc.
	mov al,'.'
ELZ9:	cmp al,7fh
	jb ELZ7
	cmp al,0AFh
	ja ELZ7
	mov al,'.'
ELZ7:	cmp bl,byte [columns]		;screen width
	jae DispEndLine			;continue reading line until end
	inc edx				;also count hidden chars (left margin)
	cmp edx,[zloffst]
	jbe ELZ5			;load new char (but no display)
	stosB
%ifdef CURSORMGNT
	clc
	call SetInverseStatus
%endif
	inc bl				;counts displayed chars only
ELZ5:	jmp DispCharLoop
;-------
FillLine:push ecx			;continue rest of line
	mov ecx,[columns]		;width
	sub cl,bl
	mov al,SPACECHAR		;fill with blanks
	jecxz FillLine2
	cmp byte[inverse],1		;special cursor attribute?
	jnz FillLine1
	stosB				;only 1st char with special attribute
%ifdef CURSORMGNT
	clc
	call SetInverseStatus
%endif
	dec ecx				;one char less
	jz FillLine2
FillLine1:
	rep stosB			;store the rest blanks
FillLine2:pop ecx
	mov byte[edi],0
	call ScreenLineShow
	dec ecx
	jnz near DispNewLine
	mov edi,[fileptr]		;restore text pointer
	jmp RestKursPos
;----------------------------------------------------------------------
; three helper subroutines called by DispNewScreen
; dealing ESC sequences for character attributes
;
SetInverseStatus:push ecx		;returns zero flag
	push esi
	jnc SIS1
	cmp byte [insstat],1
	stc
	jnz SIS4
	mov byte[inverse],1
	mov esi,reversevideoX
	add esi,[revvoff]		;switch between esc seq for linux or Xterm
	jmp short SIS2
SIS1:	cmp byte[inverse],1
	jnz SIS3
	mov byte[inverse],0
SIS6:	mov byte[isbold],0
SIS5:	mov esi,bold0
SIS2:	push byte boldlen
	pop ecx
	rep movsb
SIS3:	clc
SIS4:	pop esi
	pop ecx
	ret
;-------
SetColor:push ecx			;expects cy flag:bold /  nc:normal
	push esi
	call IsShowBlock
	jnc SCEsc1
	cmp byte [isbold],1		;never set bold if it is already bold
	jz SIS4
	mov byte [isbold],1
SCEsc2:	mov esi,bold1
	jmp short SIS2
SCEsc1:	cmp byte [isbold],0		;ditto
	jz SIS4
	jmp short SIS6
;-------
%ifdef LESSWRITEOPS
SetColor2:push ecx
	push esi
	call IsShowBlock
	jnc SIS5
	jmp short SCEsc2
%endif
;
;-------
; a little helper for SetColor* functions
;
IsShowBlock:cmp byte [showblock],0
	je SBlock
	cmp dword [blockbegin],0
	je SBlock
	cmp [blockbegin],esi
	ja SBlock
	cmp esi,[blockende]
	jb SB_ret
SBlock:	clc
SB_ret:	ret
;-------
; this helper for DispNewScreen checks screen size before writing on screen
; FIXME: adjusting edit screen resize works with xterm, but not with SVGATextMode
;
GetEditScreenSize:
%ifdef BEOS
	mov ecx,800Ch			;TCGETA+13
%else
%ifdef LINUX
	mov ecx,5413h			;asm/ioctls.h:	#define TIOCGWINSZ 0x5413
%else
	mov ecx,40087468h
%endif
%endif
	mov edx,winsize
	call IOctlTerminal
	mov eax,[edx]			;each 16 bit lines,columns
	cmp eax,0x0000FFFF		;some give no columns info..?
	jb iserr
 	or eax,eax
 	jnz noerr
iserr: 	mov eax,0x00500018		;i.e. (80<<16)+24  (assume 80x24)
noerr:	dec eax				;without status line ('dec al' are 2 byte!)
	mov byte [lines],al
	shr eax,16
	mov byte [columns],al		;columns > 255 are ignored...
	ret
;----------------------------------------------------------------------
;
; LOWER LEVEL screen acces function (main +2 helpers)
; this function does write the line buffer to screen i.e. terminal
;
; at first 2 special entry points:
WriteTwo:mov [screenline],ecx
StatusLineShow:xor ecx,ecx		;0 for last line
;
ScreenLineShow:pusha			;expecting in ecx screen line counted from 0
%ifdef LESSWRITEOPS
	mov eax,[columns]
	lea eax,[eax+32]		;estimated max ESC sequences extra bytes (i.e. boldlen*X) (BTW add eax,32 islonger)
	mul ecx				;setting edx to 0
	mov ebx,edx			;flag
	lea edi,[eax+screenbuffer]
%else
	xor edx,edx
%endif
	cld
	mov esi,screenline
sl3:	lodsb
	inc edx				;count message length to write
%ifdef LESSWRITEOPS
	cmp edi,screenbuffer_end	;never read/write beyond buffer
	jnb sl5
	cmp al,[edi]
	jz sl4
	mov [edi],al
sl5:	inc ebx				;set flag whether line need redrawing
sl4:	inc edi
%endif
	or al,al
	jnz sl3
	dec edx				;one too much
%ifdef LESSWRITEOPS
	or ebx,ebx			;redraw ?
	jz NoWrite			;..no: quick exit
%endif
	push edx
	xor edx,edx
	mov dh,byte [lines]
	sub dh,cl
	call sys_writeKP		;set cursor pos before writing the line
	pop edx
	push ecx
	mov eax,screencolors1		;set bold yellow on blue
	call sys_writeSLColors		;special for status line (ecx==0)
	mov ecx,screenline		;second argument: pointer to message to write
	call WriteFile0
	pop ecx
	mov eax,screencolors0		;reset to b/w
	call sys_writeSLColors		;special for status line (ecx==0)
	mov edx,[kurspos2]
	call sys_writeKP		;restore old cursor pos
NoWrite:popa
	ret
;-------
; a helper for ScreenLineShow
;
sys_writeSLColors:
	jecxz syswSL			;do nothing if not in status line
	ret
syswSL:	pusha
	xchg eax,ecx			;parameter points to ESC-xxx color string
	push byte scolorslen
	pop edx
	call WriteFile0
	popa
	ret
;----------------------------------------------------------------------
;
; getting line INPUT from terminal / UNDER CONSTRUCTION
;
; expecting pointer to message text in esi
;
InputStringWithMessage:
	call WriteMess9MakeLine
	mov ecx,optbuffer
	push byte optslen
	pop edx
	jmp short InputString
InputString0:call WriteMess9MakeLine
	mov edx,maxfilenamelen
;
; expecting input line buffer in ecx
; expecting max count byte in edx
; return length in eax, CY for empty string (or user abort)
;
InputString:push ecx
	push edi
	push byte 2
	pop eax
	xchg eax, [VICmdMode]
	push eax			;LONGER: push dword [VICmdMode], mov byte [VICmdMode],2
	push dword [kurspos2]
	mov ebx,[columns]
	sub ebx,stdtxtlen
	cmp edx,ebx			;TODO enable some scrolling:
	jb IS8				;not yet ready, so truncate at end of line
	mov edx,ebx
IS8:	xor ebx,ebx
	mov edi,ecx
IS0:	push ebx
	push edx
	mov byte [kurspos2],bl		;colum
	add byte [kurspos2],stdtxtlen	;offset
	mov bl,byte[lines]		;line#
	mov byte [kurspos2+1],bl
%ifdef LESSWRITEOPS
	mov byte [screenbuffer],0	;switching off usage of buffer v0.7
%endif
	call StatusLineShow
	call GetChar
	pop edx
	pop ebx
	cld
;-------
	cmp byte[mode],VI
	jnz NO_VI01
	cmp al,0
	je ISA
NO_VI01:
	call CheckUserAbort
	jne IS9
ISA:	xor ebx,ebx			;length 0 triggers CY flag
	jmp short IS1
IS9:	cmp al,RETURN
	je IS1
	cmp al,8			;^H (translated DEL)
	jne IS2
	or ebx,ebx			;@left border?
	jz IS0
	dec ebx
	dec edi
	mov al,SPACECHAR
	mov byte [ebx+screenline+stdtxtlen],al
	jmp short IS0
;-------
IS2:	cmp al,SPACECHAR
	jb IS0
	stosb
	mov byte [ebx+screenline+stdtxtlen],al		;ditto
	inc ebx
	cmp ebx,edx
	jb IS0
;-------
IS1:	xor eax,eax
	stosb				;make asciz string
	pop dword [kurspos2]
	pop dword [VICmdMode]		;restore original vi mode
	pop edi
	pop ecx
	xchg eax,ebx
	cmp al,1			;set cy flag if empty string (len always <256)
	ret				;eax length (but is < 255)
;----------
;
; GetChar (main function for kbd input)
;
ReadChar:mov eax,edi
	xchg eax,[old] 			;for ^QP
	mov [veryold],eax
GetChar:call ReadOneChar		;ah=0xFF
	cmp al,7FH
	jne RC_No7F			;special case: remap DEL to Ctrl-H
%ifndef FREEBSD
	mov al,8
%else
	mov al,7
%endif
RC_No7F:
;-------
%define DoNo 10
;-------
;
; vi needs special handling of the ESC key
;
	cmp byte [mode],VI
	jz ISVI7
	cmp al,27 			;ESC ?
	jnz near RCready_2
	call ReadOneChar		;dont care whether '[' or 'O' (should be [ for vt220 family  O for vt100 family)
	jmp short NOVI7
;-------
ISVI7:	cmp byte [VICmdMode],1
	jne NoCMDmode
	cmp al,27
	je ESCpressed
	cmp al,VIsize
	ja near Other
	mov ebx,VIcmdTable		;process command mode keys......
	jmp RCready_0			;....and ready
;-------
ESCpressed:call ReadOneChar
	cmp al,'['			;decide: it's a cursor key?
	je near Other			;yes, contine
	jmp short NoCursorKey		;no push back char into buffer and exit
NoCMDmode:cmp al,27 			;ESC ?
	jnz near RCready_2
	call KeyVImode1			;ESC pressed in EDIT Mode
	call RestoreStatusLine		;immediate update of status line (maybe bad?)
	call ReadOneChar
	cmp al,'['			;starting sequence of cursor key found?
	je IsCursorKey			;pressed ESC, but do _NOT_ switch init cmd mode
NoCursorKey:mov byte [VIbufch],al	;push char back into read buffer due it's not a cursor key
	mov al,DoNo			;do nothing
	jmp short JmpRCready
IsCursorKey:call KeyVImode0		;reset mode to 'no_command' and continue
;-------
NOVI7:	cmp byte [mode],NE		;ALT keys are currently used for ndit mode...
	jnz NONE7
	cmp al,'i'
	jnz NOi
	mov al,0x10
	jmp short JmpRCready
NOi:	cmp al,'I'
	jnz NONE7
	mov al,0x10
	jmp short JmpRCready
NONE7:	cmp byte [mode],EM
	jnz NOEM7			;ALT keys are currently used for Emacs mode...
	cmp al,'%'			;...except altH for online Help
	jne NoAltPer
	mov al,0x28
JmpRCready:jmp short RCready_1
NoAltPer:cmp al,'<'
	jne NoAltLt
	mov al,0x0e
	jmp short RCready_1
NoAltLt:cmp al,'>'
	jne NoAltGt
	mov al,0x0f
	jmp short RCready_1
NoAltGt:and al,0x5F			;to upper case
	sub al,'B'			;1at in table
	js Other
	cmp al,ATsize
	ja Other
	mov ebx,EmaAltTable
	jmp short RCready_0
NOEM7:	and al,0x5F
	cmp al,'H'
	jnz Other
	mov al,0x3D
	jmp short RCready_1
;-------
Other:	call ReadOneChar
	cmp al,'8'
	ja NoNumber
	push eax			;0,1,2....8  (i.e. 9 keys)
	call ReadOneChar
	xchg eax,ebx
	pop eax
	cmp bl,'~'			;if it's a number we expect following a '~'
	jne near GetChar
NoNumber:sub al,'0'
	cmp al,9
	jb IsNumber
	sub al,('A'-'0'-9)
	cmp al,9
	jb near GetChar
	cmp al,STsize
	ja near GetChar
IsNumber:mov ebx,ScanTable
RCready_0:xlatb
RCready_1:shl eax,8			;shift into ah (ah always != 0xFF)
RCready_2:ret
;-------
; called by ReadChar/GetChar
;
ReadOneChar:cmp byte [mode],VI
	jnz NOVI4
	xor eax,eax
	xchg eax,[VIbufch]		;preread char in buf?
	or eax,eax
	jne RoneC
NOVI4:	mov ecx,read_b			;pointer to buf
	xor edx,edx
	inc edx				;mov edx,1  (length)
	call ReadFile0
	mov eax,[ecx]			;[read_b]
RoneC:	mov ah,0xFF
	ret
;----------------------------------------------------------------------
;
; L O O K functions
; search special text locations and set register edi to
;
LookBackward:				;set EDI to 1 before EOL (0Ah) i.e., 2 before start of next line
	push ecx
	push ebx
	xor ebx,ebx
	cmp byte[edi-1],NEWLINE		;at BOL ?
	jz LBa3
	cmp byte[edi],NEWLINE		;at EOL ?
	jnz LBa1
	dec edi				;at EOL ? start search 1 char earlier
	inc ebx				;increase counter
LBa1:	mov ecx,99999
	mov al,NEWLINE
	std
	repne scasb
	lea eax,[ebx+99997]		;mov eax,99997 / add eax,ebx
	sub eax,ecx
LBa5:	pop ebx
	pop ecx
	jmp short CheckBof
;-------
LBa3:	xor eax,eax
	dec edi
	dec edi
	jmp short LBa5
;-------
LookForward:push ecx			;don't touch edx (if called by BZNLoop only)
	mov ecx,99999
	mov al,NEWLINE
	cld
	repne scasb
	mov eax,99998
	sub eax,ecx
	pop ecx
	dec edi
CheckEof:cmp edi,ebp			;ptr is eof-ptr?
	jnz CheckEnd			;Z flag if eof
	jmp short CheckENum
CheckBof:cmp edi, sot-1
	ja CheckEnd
CheckENum:mov byte [numeriere],1	;if bof
CheckEnd:ret
;-------
LookPgBegin:mov edx,[kurspos2]		;called by DispNewScreen to get sync with 1st char on screen
	movzx ecx,dh			;called by KeyCtrlQE  (go upper left)   OLD: xor ecx,ecx mov cl,dh
	inc ecx				;'inc cl' are 2 Bytes
	jmp short LookPU2
;-------
LookPgEnd:mov edx,[kurspos2]		;goes 1st char last line on screen
	mov ecx,[lines]
	sub cl,dh
        jmp short LookPD2
;-------
LookLineUp:push byte 2			;2 lines: THIS line and line BEFORE
	pop ecx
	dec dword [linenr]
	jmp short LookPU2
;-------
LookLineDown:push byte 2		;2 lines: THIS and NEXT line
	pop ecx
	inc dword [linenr]
	jmp short LookPD2
;-------
LookPageUp:mov ecx,[lines]
	dec ecx				;PgUp,PgDown one line less
LookPU1:sub [linenr],ecx
	inc ecx
LookPU2:call LookBackward
	inc edi				;inc keeps CY flag!
	jb LookPUEnd			;if BOF
	loop LookPU2			;after loop edi points to char left of 0ah
LookPUEnd:inc edi			;now points to 1st char on screen or line
	ret
;-------
LookScrDn:xor ecx,ecx
	inc ecx
	jmp short LookPD1
LookScrUp:xor ecx,ecx
	inc ecx
	jmp short LookPU1
LookHalfPgUp:mov ecx,[lines]		;vi special
	dec ecx
	shr ecx,1
	jmp short LookPU1
LookHalfPgDn:mov ecx,[lines]
	dec ecx
	shr ecx,1
	jmp short LookPD1
;-------
LookPgDown:mov ecx,[lines]
	dec ecx				;PgUp,PgDown one line less
LookPD1:add [linenr],ecx
	inc ecx
LookPD2:call LookForward
	jz LookPDEnd			;(jmp if EOF)
	inc edi				;1st char next line
	loop LookPD2
	dec edi				;last char last line
LookPDEnd:sub edi,eax			;1st char last line
	ret
;----------------------------------------------------------------------
;
; some more CHECK functions
;
CheckBlock:cmp byte [showblock],1	;returns CY if error else ok: NC
	jc CheckBlockEnd
	mov esi,[blockende]
	cmp esi, sot
	jb CheckBlockEnd
	mov esi,[blockbegin]		;side effect esi points to block begin
	cmp esi, sot
	jb CheckBlockEnd
	cmp [blockende],esi     	;^KK > ^KB ..OK if above!
CheckBlockEnd:ret
;-------
CheckImBlock:cmp [blockbegin],edi	;^KB mark > edi ?
	ja CImBlockEnd			;OK
	cmp edi,[blockende]		;edi > ^KK
CImBlockEnd:ret	          		;output:cy error / nc ok inside block
;-------
CheckMode:cmp byte [edi],NEWLINE	;checks for INSERT status
	jz ChModeEnd
	cmp byte [insstat],1
ChModeEnd:ret				;Z flag for ins-mode
;-------
; a special case called by DeleteByteCheckMarker
;
CheckMarker:				;edx is blockbegin (^KB)
					;ebx is deleate area end --- edi delete area start
	cmp edi,edx			;delete area start < ^KB marker ?
	ja CMEnd			;no
	cmp ebx,edx			;yes, but delete area end > ^KB ?
	jb CMEnd			;no
	mov edx,edi			;yes so block start (^KB) to delete area start
CMEnd:	ret
;----------------------------------------------------------------------
;
; C O U N T  functions
; to return number of chars up to some place
; (all of them are wrappers of Look....functions anyway)
;
CountToLineEnd:push edi
	call LookForward
	pop edi
	ret				;eax=chars up to line end
;-------
CountColToLineBeginVis:			;counts columns represented by chars in EAX
	call CountToLineBegin		;i.e. EXPAND any TAB chars found
	push esi
	xor edx,edx
	mov esi,edi			;startpoint
	sub esi,eax			;to bol
	dec esi
CCV1:	inc esi
	cmp esi,edi
	jae CCVend
	cmp byte [esi],TABCHAR
	jz CCVTab
	inc edx				;count visible chars
	jmp short CCV1
CCVTab:	call SpacesForTab		;return space_up_to_next_tab in ah
	add dl,ah			;FIXME: now using 8 bits only
	jmp short CCV1
CCVend: mov [ch2linebeg],edx		;ch2linebeg: interface to Key... functions
	mov eax,edx			;eax: interface to DispNewScreen
	pop esi
	ret
;-------
CountToLineBegin:push edi		;output eax=chars up there
	call LookBackward
	mov esi,edi			;side effect: set edi to 1st char in line
	pop edi
	ret
;-------
CountToWordBegin:			;output eax=chars up there
	mov esi,edi
CountNLoop:inc esi
	cmp byte [esi],NEWLINE
	jz CTWend
	cmp byte [esi],SPACECHAR	;below SPACE includes tab chars
	jbe CountNLoop
	cmp byte [esi-1],2fh
	ja CountNLoop
CTWend:	mov eax,esi
	sub eax,edi			;maybe =0
Goret:	ret
;----------------------------------------------------------------------
;
; Online Help: show the message followed by common text
;
KeyHelp:push dword [kurspos]
	pusha
	xor eax,eax
	xchg eax,[showblock]		;hide an blockmarker
	push eax
	cld
	mov esi,sot			;save "buffercopysize" of text
	mov edi,buffercopy
	mov ecx,buffercopysize
	push edi
	push ecx
	push esi
;-------
	rep movsb
	call GetHelpText
	pop edi
	push edi			;i.e. mov edi,sot
;-------
	push edi
	rep movsb			;overwrite saved text with help message
	mov esi,helpfoot
	mov ecx,helpfootsize
	rep movsb
	mov ebp,edi			;set END_OF_HELP_TEXT pointer
	pop edi
	call DispNewScreen
	call ReadOneChar		;wait for a pressed char
;-------
	pop edi
	pop ecx
	pop esi				;former edi
	cld
	rep movsb			;restore textbuffer with saved patr
	pop dword [showblock]
	popa
	pop edx				;cursor pos
	jmp short SetKursPos
;---------------------------------------------------------------------
;
; some CURSOR control functions
;
GoUp:	xor eax,eax
	jmp short UpDown
GoDown:	mov al,byte [lines]
	dec eax				;'dec al' are 2 byte!
	mov ah,-1
UpDown:	mov edx,[kurspos2]		;former was call getkurspos
	cmp al,dh
	jz Goret
	sbb dh,ah			;ONLY here we change curent line of cursor
	jmp short SetKursPos
;-------
; set cursor to some desired places
;
KeyEmaCtrlL:call CountToLineBegin
	mov dh,byte [lines]		;move cursor to center line (and later redisplay)
	shr dh,1
	mov dl,al
	jmp short SetKursPos
KursorFirstLine:xor edx,edx
	jmp short SetKursPos
KursorLastLine:mov dh,byte [lines]
	dec dh
	mov dl,0
	jmp short SetKursPos
KursorStatusLine:mov dh,byte [lines]
	mov dl,stdtxtlen
	jmp short SetKursPos
RestKursPos:mov edx,[kurspos]
SetKursPos:mov [kurspos2],edx      	;saves reading cursor pos   (0,0)
sys_writeKP:pusha
	call make_KPstr
	mov ecx,setkp 			;second argument: pointer to message to write
	push byte setkplen		;third argument: message length
	pop edx
	call WriteFile0
	popa
	ret
;-------
; make ESC sequence appropriate to most important terminals
;
make_KPstr:inc dl			;expecting cursor pos in dh/dl (0,0)
	inc dh 				;both line (dh) col (dl) are counted now from 1
	cld
	mov edi,setkp			;build cursor control esc string db 27,'[000;000H'
	mov al,1Bh
	stosb				;init memory
	mov eax,'[000'
	stosd
	mov al,';'			;i.e. load eax with ';000'
	stosd
	mov al,'H'
	stosb
	mov edi,setkp+1+3		;line end
	movzx eax,dh			;DH=line
	push edx
	call IntegerToAscii		;make number string
	pop edx
	mov edi,setkp+1+3+4		;column end
	movzx eax,dl			;DL=col
;-------continued...
; a general helper
;   expects integer# in eax
IntegerToAscii:
	or eax,eax
	jns ItoA1
	;int 3				;Assertation
	xor eax,eax			;this should never be
	inc eax
ItoA1:	push byte 10
	pop ecx
	std
	xchg eax,ebx			;ebx helper (xchg eax,.. is only 1 byte!)
Connum1:xchg eax,ebx
	cdq
	div ecx
	xchg eax,ebx			;save quotient (new low word)
	mov al,dl
	and al,0fh
	add al,'0'
	stosb
	or ebx,ebx
	jne Connum1
	cld
	ret
;----------------------------------------------------------------------
;
; functions for INSERTING, COPYING and DELETING chars in text
;
InsertByte:or eax,eax			;input: eax = # of bytes  edi = ptr where
	jz near Ins3
	mov ecx,[maxlen]		;max_len+offset-eofptr=freespace(ecx)
	add ecx,sot
	sub ecx,ebp
	mov edx,[EmaKiSize]
	sub ecx,edx			;sub size of kill buffer from free space
	cmp ecx,eax			;cmp freespace - newbytes  ;>= 0 ok/ NC  <0 bad / CY
	jnc SpaceAva
	push byte 105
	pop dword [errno]		;(mov dword[errno],105 has 2 byte extra)
	call OSerror
	call RestKursPos
	stc
	ret
SpaceAva:push edi
	mov esi,ebp			;end of text movsb-source
	lea ecx,[ebp+1]
	sub ecx,edi			;space count: distance between eof and current position
	lea edi,[ebp+eax]		;movsb-destination
	cmp byte [mode],WS
	jz ISWS8
	add ecx,edx			;add size of kill buffer to distance
	add edi,edx
	add esi,edx
ISWS8:	std
	rep movsB
Ins0:	pop edi				;here is the jmp destination from DeleteByte
;-------
	mov byte [changed],CHANGED
	add ebp,eax
	cmp byte [mode],WS
	jnz NOWS8
	cmp edi,[blockende]
	jae Ins1
	add [blockende],eax
Ins1:	cmp edi,[blockbegin]
	jae Ins2
	add [blockbegin],eax
NOWS8:
	test byte [mode], EM | PI
	jz NO_EM02
	cmp edi,[EmaMark]
	jae Ins2
	add [EmaMark],eax
NO_EM02:
Ins2:	clc
Ins3:	ret				;output:nc=ok/cy=bad /ecx=0/ eax inserted / -eax deleted
;-------
DeleteByteCheckMarker:			;edi points to begin
	cmp byte [mode],WS
	jnz DeleteByte
	lea ebx,[edi+eax]		;ebx points to end
	mov edx,[blockbegin]
	call CheckMarker
	mov [blockbegin],edx
	mov edx,[blockende]
	call CheckMarker
	mov [blockende],edx
DeleteByte:or eax,eax			;input in eax
	jz MoveBlEnd
	push edi
	mov ecx,ebp			;end
	sub ecx,edi
	lea esi,[edi+eax]		;current + x chars
	sub ecx,eax
	cmp byte [mode],WS
	jz No_WS8
	add ecx,[EmaKiSize]
No_WS8:	inc ecx
	cld
	rep movsb
	neg eax				;"neg eax" is for continuing @InsertByte
	jmp Ins0			;pending "pop edi"
;-------
CopyBlock:call CheckBlock		;copy block, called by ^KC, ^KV
	jc MoveBlEnd
	call CheckImBlock
	jc MoveBlEnd
	mov eax,[blockende]
	sub eax,esi			;block len
	call InsertByte
	jc MoveBlEnd
	mov esi,[blockbegin]
;-------
MoveBlock:push edi			;input : esi=^KB edi=current
	mov ecx,eax			;don't use xchg here
	cld
	rep movsb
	pop edi
	clc				;nocarry->ok
MoveBlEnd:ret				;return eax=size
;----------------------------------------------------------------------
KeyVICmdyy:push edi
	call KeyHome
	mov byte [showblock],1
	mov [EmaMark],edi
	call KeyEnd
	inc edi				;add a line delimiter
	call KeyEmaAltW
	pop edi
	ret
;-------
KeyVICmdy:call ReadOneChar
	cmp al,'y'
	je KeyVICmdyy
	cmp al,"'"
	jne MoveBlEnd
	call ReadOneChar
	cmp al,'a'			;y'a
	jne MoveBlEnd
	mov ecx,[EmaMark]		;don't go forther if no mark set
	jecxz MoveBlEnd
	mov ecx,[showblock]
	jecxz MoveBlEnd
	call KeyEnd
	inc edi				;add a line delimiter
	call KeyEmaAltW
	mov edi,[EmaMark]
	jmp CQFNum
;
; some of the EM specials
;
KeyEmaAltW2:pusha
	mov edi,ebp
	inc edi
	cmp byte [mode],VI
	jz KEW
	cmp byte [EmaCtrlK],1
	jnz KEW
	add edi,[EmaKiSize]
	add [EmaKiSize],eax
	jmp short KE2
KEW:	mov [EmaKiSize],eax
	mov [EmaKiSrc],esi
KE2:	mov ecx,eax
	cld
	rep movsb
	mov byte [showblock],cl		;cl==0
Keaw2:	mov byte [EmaCtrlK],1
	popa
	ret
;-------
KeyEmaAltW:mov byte [VInolinebased],0
	pusha
	mov ecx,[showblock]
	jecxz Keaw2
	mov ecx,[EmaMark]
	jecxz Keaw2
	mov eax,edi
	cmp ecx,eax
	jb KEAW
	xchg eax,ecx
KEAW:	sub eax,ecx			;eax end / ecx beg
	mov esi,ecx
	mov edi,ebp
	inc edi
	mov [EmaKiSize],eax
	mov [EmaKiSrc],esi
	xchg eax,ecx			;OLD mov ecx,eax 1 byte longer
	cld
	rep movsb
	mov byte [showblock],cl		;cl==0
	popa
KeaWRet:ret
;-------
KeyEmaCtrlY:mov ecx,[EmaKiSize]
	jecxz KeaWRet
	xchg eax,ecx			;OLD mov eax,ecx 1 byte longer
	push eax
	call InsertByte
	pop ecx
	jc KeaWRet			;no_space_error is handled in InsertByte
	mov esi,ebp
	inc esi
	mov [EmaMark],edi
	cld
	rep movsb
	mov byte [showblock],cl		;cl==0
	cmp byte [mode],VI
	jz ISVI9
	call KeyEmaCtrlL
ISVI9:	jmp CQFNum
;-------
KeyEmaCtrlW:mov ecx,[showblock]
	cmp byte[mode],PI
	jne NOPI1
	jecxz DelLine
	mov ecx,[EmaMark]
	jecxz DelLine
	jmp short NOPI2
NOPI1:	jecxz KeaWRet
	mov ecx,[EmaMark]
	jecxz KeaWRet
NOPI2:	call KeyEmaAltW
	mov edi,[EmaKiSrc]
	mov eax,[EmaKiSize]
	jmp DeleteByte
DelLine:jmp KeyCtrlY

;----------------------------------------------------------------------
;
; functions reading/writing  text or blocks  from/into  files
;
KeyVICmdE:lea esi,[ecx+2]
	cmp byte [esi],SPACECHAR
	je KeaWRet
	pusha				;save before load new
	call SaveFile
	popa
;-------continue
NewFile:cld
	call InitVars
	or esi,esi
	jz NFnoarg
	mov edi,filepath
NF1:	lodsb
	stosb
	or al,al
	jnz NF1
	jmp short GetFile
;-------
NFnoarg:mov esi,helptext		;initial part of help
	mov edi,sot
	push byte helptextsize
	pop ecx
	push edi
	rep movsb
;-------
	call GetHelpText		;second part of help
	rep movsb
	pop edi
	lea ebp,[sot+helptextsize+help_ws_size]
	call DispNewScreen
;-------
	mov esi, filename
	mov ecx, filepath
	call InputString0
	jc near NFEnd2			;empty string not allowed here
;-------
GetFile:
%ifdef BEOS
	xor ebx,ebx
	mov edx,ebx
	dec ebx				;edx==0,ebx==-1
	mov ecx,filepath
%else
	mov ebx,filepath
%endif
	call OpenFile0
	mov edi,sot
	mov ebp,edi
	js NewFileEnd
%ifdef LINUX
	call Seek
	pusha
	lea ebx,[eax+eax+102400]	;twice filesize plus reserve = space for inserts
	mov [maxlen],ebx
	add ebx,text
	call SysBrk
	popa
	js OSejmp1			;OSerror
%else
	mov ebx,eax			;for FreeBSD memory is hard coded by maxlen
%endif
;-------
	call Fstat
	js OSejmp1			;OSerror
	mov eax,[fstatbuf+8]		;FIXME: define structure
	and eax,777q
	mov [perms],eax
%ifdef LINUX
	mov eax,[fstatbuf+40]		;FIXME: define structure
	mov [changetime],eax
	mov eax,[fstatbuf+12]		;FIXME: define structure
	mov [giduid],eax
%else
%ifdef FREEBSD
	mov eax,[fstatbuf+32]		;FIXME: define structure
	mov [changetime],eax
	mov eax,[fstatbuf+12]
	mov [uid],eax
	mov eax,[fstatbuf+16]
	mov [gid],eax
%endif
%endif
OldFile1:mov edx,[maxlen]		;i.e. either 'max' or filesize twice
	mov ecx,edi			;sot
	call ReadFile
	xchg edx,eax			;mov edx,eax	bytes read
	js OSejmp0			;OSerror
	call CloseFile
OSejmp1:js OSejmp0			;OSerror
	lea ebp,[edx+sot]		;eof_ptr=filesize+start_of_text
NewFileEnd:mov byte [ebp],NEWLINE	;eof-marker
	clc
NFEnd2:	ret
;-------
;  save file (called by ^KS,^KX)
;
SaveFile:cmp byte [changed],UNCHANGED
	jz NFEnd2			;no changes: nothing to save
	mov esi, filesave
	call WriteMess9
;-------
	mov esi,filepath
	push edi
	mov edi,bakpath
	mov ebx,esi
	mov ecx,edi
	cld
SF0:	lodsb
	stosb				;copy to BAK file path
	or al,al
	jne SF0
	dec edi
	push byte '~'			;add BAK file extension
	pop eax
	stosw				;stosw not stosb because ascii-ZERO
	pop edi
%ifdef BEOS
	push edi
	mov ebx,0xFFFFFFFF
	mov edx,ebx
	mov ecx,filepath
	mov edi,bakpath
%endif
%ifdef MAKE_BACKUP
	cmp dword [ecx],'/tmp'
	je no_ren
%ifdef BEOS
	call RenameFile                 ;ecx is filepath
%else
	pusha
	mov ecx,linkbuffer		;we are only interested whether symlink or not
	push byte linkbuffersize	;=4 byte dummy buffer
	pop edx
	call ReadLink
	popa
	jns CopyBAK			;no error means it's a symlink...
	call RenameFile                 ;...but plain files are easy to rename (ecx is filepath)
	jmp short no_ren		;...simlilar behave 'xemacs' and 'jed'
CopyBAK:call CopyToBackup		;we can't simply rename the link
%endif
no_ren:	;...continue here...
%endif
%ifdef BEOS
	pop edi
%endif
;-------
%ifdef BEOS
	mov ebx,0xFFFFFFFF
	mov ecx,filepath
	mov edx,0x777
%else
	mov ecx,O_WRONLY_CREAT_TRUNC
	mov edx,[perms]
%endif
	call OpenFile
OSejmp0:js OSejmp9			;OSerror
	xchg ebx,eax			;file descriptor  (xchg is only 1 byte)
;-------
%ifdef LINUX
	mov ecx,[giduid]
	mov edx,ecx
	shr edx,16
	and ecx,0xffff
	call ChownFile
%else
%ifdef FREEBSD
	mov edx,[gid]
	mov ecx,[uid]
	call ChownFile
%endif
%endif
;-------
	mov ecx,sot			;ecx=bof
	mov edx,ebp			;eof
SaveFile2:sub edx,ecx			;edx=fileesize= eof-bof
	cmp byte[mode],VI
	jnz NoAddNL
	cmp byte [ebp-1],NEWLINE
	jz NoAddNL
	inc edx				;append NewLine char for VI mode
NoAddNL:call WriteFile			;ebx=file descriptor
OSejmp9:js OSejmp			;OSerror
	push byte 5
	pop dword[errno]		;just in case of....
	cmp eax,edx			;all written?
	jnz near OSerror
	call CloseFile
	js OSejmp			;OSerror
SaveFile3:ret
;-------
;  save block (called by ^KW)
;
SaveBlock:call GetBlockName
	jc near DE2
SaveBl2:
%ifdef BEOS
	mov ebx,0xFFFFFFFF
	mov ecx,blockpath
	mov edx,0x777
%else
	mov ecx,O_WRONLY_CREAT_TRUNC
	mov ebx,blockpath
	mov edx,PERMS
%endif
	call OpenFile
	js OSejmp			;OSerror
	mov ecx,esi			;= block begin
	mov edx,[blockende]
	xchg ebx,eax			;file descriptor  (xchg is only 1 byte)
	jmp short SaveFile2
;-------
; read a block into buffer (by ^KR)
;
ReadBlock:call GetBlockName
	jc near DE2
ReadBlock2:
%ifdef BEOS
	xor ebx,ebx
	mov edx,ebx
	dec ebx				;edx==0,ebx==-1
	mov ecx,blockpath
%else
	mov ebx,blockpath
%endif
	call OpenFile0
OSejmp:	js OSerror
	call Seek
	js OSerror
	push eax			;eax=fileesize
	call InsertByte
	pop edx				;file size
	jc SaveFile3			;ret if cy InsertByte told an error message itself
	mov ecx,edi			;^offset akt ptr
	call ReadFile
	js preOSerror			;to delete inserted bytes (# in EDX)
	mov ecx,eax    			;bytes read
	call CloseFile
	js OSerror
	mov word[errno],5		;just in case of....
	cmp edx,ecx			;all read?
	jnz OSerror
	ret
;-------
preOSerror:mov eax,edx			;count bytes
	call DeleteByte			;delete space reserved for insertation
OSerror:push edi
	mov edi,error+8			;where to store ASCII value of errno
	mov eax,[errno]
	push eax
	call IntegerToAscii
	pop ecx				;for getting error text via LookPD2
	mov edi,errmsgs
	call LookPD2			;look message # (ecx) in line number #
	mov esi,edi
	mov edi,error+9
	mov ax,' :'
	stosw
	push byte 80			;max strlen / compare errlen equ 100
	pop ecx
	rep movsb
DE0:	mov esi,error
	pop edi
DE1:	call WriteMess9
	call ReadOneChar		;was GetChar
DE2:	;continued...
;----------------------------------------------------------------------
;
; more STATUS LINE maintaining subroutines
;
RestoreStatusLine:pusha			;important e.g. for asksave
	call InitStatusLine
	mov ecx,[columns]		;width
	cmp cl,stdtxtlen+15+5+2		;this window is too small
	jb near RSno_lineNr
	mov bl,byte[changed]
	mov byte[screenline+1],bl	;changed status
;-------
	cmp byte [mode],VI		;vi doesn't get help text because AltH won't work...
	jz IsVI00			;...caused by different handling due single ESC keys
%ifndef LINUX
	lea eax,[ecx-13+screenline]	;FreeBSD or Beos
%else
	lea eax,[ecx-12+screenline]
%endif
	mov dword [eax],'altH'
	lea eax,[eax+4]
	mov dword [eax],'=hel'		;'p' is stored with editor mode name
	mov ebx,editmode
	mov edx,[ebx]
	cmp byte [mode],PI
	jnz No_PI1
	mov edx,[ebx+4]
No_PI1:	cmp byte [mode],EM
	jnz No_Em1
	mov edx,[ebx+8]
No_Em1:	cmp byte [mode],NE
	jnz No_Ne1
	mov edx,[ebx+16]
No_Ne1:	lea eax,[eax+4]
	mov [eax],edx
	jmp short NOVI0
;-------
IsVI00:	cmp byte [VICmdMode],1
	jnz NOVI0
	mov ebx,'CMD '
	jmp short RSL1
NOVI0:	mov ebx,'INS '			;Insert
	cmp byte [insstat],1
	jz RSL1
	mov ebx,'OVR '			;Overwrite
RSL1:	mov [screenline+4],ebx		;mode status
	mov edi,screenline+stdtxtlen
	lea ecx,[ecx-(stdtxtlen+15+5)]	;space for other than filename
	mov esi,filepath		;lea ... is shorter than sub ecx,stdtxtlen+15+5
RSL2:	lodsb
	or al,al
	jz RSL4
	stosb
	loop RSL2
RSL4:	mov edi,screenline-15
	add edi,[columns]
	mov eax,[columne]
	inc eax				;start with 1
	call IntegerToAscii
	mov byte [edi],':'		;delimiter ROW:COL
	dec edi
	mov eax,[linenr]
	call IntegerToAscii
RSno_lineNr:call StatusLineShow		;now write all at once
	popa
	stc				;error status only important if called via OSError
	ret
;-------
;
; write an answer prompt into status line
; (with and without re-initialisation)
; expecting esi points to ASCIIZ or 0A terminated string
;
WriteMess9MakeLine:call InitStatusLine
WriteMess9:pusha
	mov edi,screenline
	cld
WriteMLoop:lodsb
	cmp al,0ah			;for error messages
	jbe WriteMEnd
	stosb
	jmp short WriteMLoop
WriteMEnd:call StatusLineShow
	popa
	jmp KursorStatusLine
;-------
; a helper for other status line functions:
; simply init an empty line
;
InitStatusLine:pusha			;simply init an empty line
	mov edi,screenline
	mov al,SPACECHAR
	mov ecx,[columns]
%ifndef LINUX
	dec ecx				;? FreeBSD
%endif
	cld
	rep stosb
	mov al,0			;prepare ASCIIZ string
	stosb
	popa
	ret
;-------
; read a file name for block operations
;
GetBlockName:pusha
	mov esi,block
	mov ecx,blockpath
	call InputString0		;cy if empty string
	pushf
	call RestKursPos
	popf
	popa
	ret
;-------
; helper for NewFile
;
InitVars:mov byte [text],NEWLINE	;don't touch esi!
	mov byte [changed],UNCHANGED
	xor eax,eax
	mov dword [oldQFpos],eax
	mov byte[bereitsges],al
	mov [blockbegin],eax
	mov [blockende],eax
	mov byte [VICmdMode],1
	mov [endeedit],al
	mov dword[old],sot
	mov byte [showblock],al
	mov [EmaMark],eax
	inc eax
	mov dword [linenr],eax
	mov byte [showblock],al
	mov byte [insstat],al
	mov dword [maxlen],max
	mov dword [error],'ERRO'
	mov dword [error+4],'R   '
	mov dword [perms],PERMS
	dec eax
	dec eax
%ifdef LINUX
	mov dword [giduid],eax		;eax == -1 i.e. no changes in fchown
%else
%ifdef FREEBSD
	mov dword [gid],eax
	mov dword [uid],eax
%endif
%endif
	ret
;-------
%ifdef BEOS
Seek:	xchg ebx,eax			;mov file_descriptor to ebx (xchg is 1 byte only)
	push byte 2
	pop edx
	call SeekFile			;end
	js SeekRet
	xor edx,edx
	push eax
	call SeekFile			;home
	pop eax
SeekRet:ret
%else
%ifdef LINUX
Seek:	xchg ebx,eax			;mov file_descriptor to ebx (xchg is 1 byte only)
	push byte 2
	pop edx
	call SeekFile			;end
	js SeekRet
	xor edx,edx
	push eax
	call SeekFile			;home
	pop eax
SeekRet:ret
%else					;64 bit
Seek:	push edi
	push esi
	mov ebx,eax			;file desc
	xor edx,edx
	xor edi,edi
	push byte 2
	pop esi
	call SeekFile			;end
	js SeekRet
	xor esi,esi
	push eax
	call SeekFile			;home
	pop eax
SeekRet:pop esi
	pop edi
	ret
%endif
%endif
;----------------------------------------------------------------------
;
; FIND/REPLACE related stuff
;
AskForReplace:mov esi,askreplace1
	mov ecx,suchtext
	call InputString0
	jc AskFor_Ex
	mov [suchlaenge],eax
	mov esi,askreplace2
	mov ecx,replacetext
	call InputString0
	jmp short GetOptions		;cy flag is allowed here 18.6.00
AskForFind:mov esi,askfind
	mov ecx,suchtext
	call InputString0
	jc AskFor_Ex
GetOptions:mov [repllaenge],eax
	test byte [mode],WS|NE
	jz GetOpt2
	mov esi,optiontext
	call InputStringWithMessage	;empty string is allowd for std options...
	call ParseOptions		;...(set in ParseOptions)
GetOpt2:clc
AskFor_Ex:jnc AFE2
	mov byte [bereitsges],0
AFE2:	pushf
	call RestKursPos
	popf
	ret
;-------
; check string for 2 possible options: C,c,B,b  (case sensitive & backward)
;
ParseOptions:push esi
	cld
	mov esi, optbuffer
	push byte 1
	pop dword[vorwarts]		;mov dword[vorwarts],1 is longer
	mov byte[grossklein],0dfh
Scan1:	lodsb
	and al,5fh			;upper case
	cmp al,'C'
	jnz notCopt
	xor byte[grossklein],20h	;result 0dfh,   2*C is like not C option
notCopt:cmp al,'B'
	jnz notBopt
	neg dword[vorwarts]		;similar 2*B is backward twice i.e. forward
notBopt:or al,al
	jnz Scan1
	pop esi
	ret
;-------
; the find subroutine itself
;
find2:	mov ebx,edi
find3:	lodsb
	or al,al			;=end?
	jz found
	cmp al,41h
	jb find7
	and al,ch
find7:	inc edi
	mov cl,byte [edi]
	cmp cl,41h
	jb find10
	and cl,ch
find10:	cmp al,cl
	jz find3
	mov edi,ebx
FindText:mov ch,[grossklein]		;ff or df
	mov esi,suchtext
	cld
	lodsb
	cmp al,41h
	jb find1
	and al,ch
find1:	add edi,[vorwarts]		;+1 or -1
	mov cl,byte [edi]
	cmp cl,41h
	jb find6
	and cl,ch
find6:	cmp al,cl
	je find2
	cmp byte[mode],PI
	jnz find_WS
	cmp edi,ebp
	jb foundEOF
	mov edi,sot
	cmp edi,[PicoSearch]
	je notfound
foundEOF:jmp short find9
find_WS:cmp edi,ebp
	ja notfound
find9:	cmp edi, sot
	jnb find1
notfound:stc
	ret
found:	mov edi,ebx
	clc				;edi points after location
	ret
;----------------------------------------------------------------------
;
; some GENERAL helper functions
;
;
; Get.....ToInteger reads integer value from keyboard (only > 0)
;
GetOctalToInteger:push byte 7		;octal base-1
	jmp short GATI2
GetAsciiToInteger:push byte 9		;decimal base-1
GATI2:	cmp byte [mode],VI
	jz ISVI8
	call InputStringWithMessage	;eax = al = length
	call AskFor_Ex			;repair status line & set cursor pos / preserve flags
ISVI8:	push byte 0
	pop esi				;preserve flags
	pop ebx				;bl == base-1
	xchg ecx,esi
	jc AIexit2
	cld
AIload:	lodsb				;eax bit 8..31 are 0
	sub al,'0'
	js AIexit
	cmp al,bl
	ja AIexit
	cmp bl,7			;if base==8
	je GATI3
	lea ecx,[ecx+4*ecx]
	lea ecx,[2*ecx+eax]		;mul 10 plus digit
	jmp short AIload
GATI3:	lea ecx,[8*ecx+eax]		;mul 8 plus digit
	jmp short AIload
AIexit:	or ecx,ecx			;ret ecx
AIexit2:ret				;CY or ZR if error
;-------
;
; SpacesForTab expects current column in edx
; returns # spaces up to next tabulated location in AH
;
SpacesForTab:push ecx
	mov eax,edx
	mov cl,TAB
	xor ah,ah
	div cl
	neg ah				;ah = modulo division
	add ah,TAB			;TAB - pos % TAB
	pop ecx
	ret
;-------
;
; GetHelpText returns ecx==size of help text / esi==help text for current edit mode
;
GetHelpText:mov eax,mode
	mov esi,help_ws
	mov ecx,help_ws_size
	cmp byte [eax],PI
	jnz NoPI1
	mov esi,help_ws+help_ws_size
NoPI1:	cmp byte [eax],EM
	jnz NoEm1
	mov esi,help_ws+2*help_ws_size
NoEm1:	cmp byte [eax],VI
	jnz NoVi1
	mov esi,help_ws+3*help_ws_size
NoVi1:	cmp byte [eax],NE
	jnz NoNe1
	mov esi,help_ws+4*help_ws_size
NoNe1:	ret
;-------
;
; Check whether user discarded his input
;
CheckUserAbort:mov esi,mode
	cmp byte[esi],WS
	jz CUAWS
	cmp byte[esi],EM
	jz CUAEM
	cmp al,3			;^C abort
	ret
CUAWS:	cmp al,15h			;^U abort
	ret
CUAEM:	cmp al,7			;^G abort
	ret
;-------
KeyEditMode:
	mov esi,modetxt
	call InputStringWithMessage	;empty string is allowd for std options...
	call RestKursPos
	mov esi,optbuffer
	xor eax,eax
	mov byte [showblock],al
	mov [EmaMark],eax
	mov [EmaKiSize],eax		;do not allow side effects
	mov [blockende],eax		;i.e. NO block valid now
;;	mov [blockbegin],eax
	mov eax,mode
;-------
SetEditMode:
	cmp dword [esi],'e3ne'
	jnz NoNe
	mov byte [eax],NE
	ret
NoNe:	cmp dword [esi],'e3em'
	jnz NoEm
	mov byte [eax],EM
	ret
NoEm:	cmp dword [esi],'e3pi'
	jnz NoPi
	mov byte [eax],PI
	ret
NoPi:	cmp dword [esi],'e3vi'
	jnz NoVi
	mov byte [eax],VI
	ret
NoVi:	cmp dword [esi],'e3ws'
	jnz modeOK
	mov byte [eax],WS
modeOK:	ret
;-------
SetTermStruc:
%ifdef BEOS
	mov ecx,8000h			;TCGETA
%else
%ifdef LINUX
	mov ecx,5401h			;TCGETS asm/ioctls.h
%else
	mov ecx,402c7413h		;TIOCGETA
%endif
%endif
	mov edx,orig
	call IOctlTerminal0
	mov esi,edx
	mov edi,termios
	mov edx,edi
	push byte termios_size		;prepare a copy of original settings
	pop ecx
	cld
	rep movsb
;-------
%ifdef BEOS
	and byte [edx+12],(~2 & ~1 & ~8);icanon off, isig (^C) off, echo off
	and byte [edx+ 1],(~4 & ~1)	;ixon off,   icrnl off
	mov ecx,8001h			;TCSETA
%else
%ifdef LINUX
	and byte [edx+12],(~2 & ~1 & ~8);icanon off, isig (^C) off, echo off
	and byte [edx+ 1],(~4 & ~1)	;ixon off,   icrnl off
	mov byte [edx+23],1		;set min=1 ->needed for gnome-terminal
	mov ecx,5402h			;TCSETS asm/ioctls.h
%else
	and word [edx+12],(~100h & ~80h & ~8h)	;icanon off, isig (^C) off, echo off
	and word [edx+ 0],(~200h & ~100h) ;ixon off,   icrnl off
	mov ecx,802c7414h		;TIOCSETA
%endif
%endif
	jmp short IOctlTerminal		;edx is termios pointer
;-------
GeneralErr:mov esp,[stackptr]		;restore stack pos for math calc
VIBeepForD:pusha
	xor edx,edx
	inc edx
	mov ecx,BeepChar
	call WriteFile0
	popa
	ret
;----------------------------------------------------------------------
;
; INTERFACE to OS kernel
; we differ between BeOS, Linux and any else (in fact only FreeBSD)
;
IOctlTerminal0:mov edx,orig
IOctlTerminal:mov ebx,stdin		;expects EDX termios or winsize structure ptr
%ifdef BEOS
	mov al,4			;4 stack places
%else
	mov al,54			;54 == the ioctl syscall no.
%endif
	jmp short IntCall		;ECX TCSETS,TCGETS,TIOCGWINSZ
;------
ReadFile0:xor ebx,ebx			;mov ebx,stdin		;file desc
ReadFile:
%ifdef BEOS
	push dword 2			;4+X? stack places
	jmp short WFile
%else
	mov al,3			;system call number (sys_read) ;return read byte EAX
	jmp short IntCall		;ebx file / ecx buffer / edx count byte
%endif
;-------
WriteFile0:xor ebx,ebx			;mov ebx,stdout		;file desc
	inc ebx				;ditto
WriteFile:
%ifdef BEOS
	mov eax,3			;4+X?  stack places
	push dword 3
WFile:	pop eax
	call IntRdWr
	nop
	nop
	nop
	nop
	ret
%else
	mov al,4			;system call number (sys_write)
	jmp short IntCall
%endif
;-------
OpenFile0:
%ifndef BEOS
	xor ecx,ecx			;i.e O_RDONLY
%endif
OpenFile:
%ifdef BEOS
	mov al,0			;5 stack places
	push edi
	mov edi,0x1A4
	call IntCall
	pop edi
	ret
%else
	mov al,5			;system call number
	jmp short IntCall		;ecx mode / ebx path / edx permissions (if create)
%endif

;-------
CloseFile:
%ifdef BEOS
	mov al,1			;1 stack place
%else
	mov al,6			;system call number (close)
%endif
	jmp short IntCall		;ebx is file desc
;-------
ReadLink:
%ifdef LINUX
	mov al,85
%else
	mov al,58
%endif
	jmp short IntCall
;-------
Fstat:
%ifdef BEOS
	xor eax,eax			;dummy
	ret
%endif
	mov ecx,fstatbuf
%ifdef LINUX
	mov al,108
%else
	mov al,189
%endif
	jmp short IntCall
;-------
RenameFile:
%ifdef BEOS
	mov al,26h			;4 stack places
	jmp short IntCall
%else
%ifdef LINUX
	mov al,38
%else
	mov al,128
%endif
	jmp short IntCall
%endif
;-------
ChownFile:
%ifdef BEOS
	xor eax,eax			;dummy
	ret
%endif
%ifdef LINUX
	mov al,95
%else
	mov al,123			;fchown
%endif
	jmp short IntCall
;-------
SysBrk:
%ifdef BEOS
	xor eax,eax			;dummy
	ret
%endif
%ifdef LINUX
	mov al,45			;system call number
%else
	mov al,69
%endif
	jmp short IntCall		;ebx addr
;-------
Exit:	xor ebx,ebx
Exit2:
%ifdef BEOS
	mov al,3fh
%else
	mov al,1
%endif
	jmp short IntCall
;-------
SeekFile:xor ecx,ecx			;ecx offset / ebx file / edx method
%ifdef BEOS
	mov al,5			;4 stack places (using 64 bit for ptr)
	push edi
	push edx
	mov edi,edx
	xor edx,edx
	call IntCall
	pop edx
	pop edi
	ret
%else
%ifdef LINUX
	mov al,19			;system call number (lseek)
%else
	xor edx,edx
	xor edi,edi
	mov al,199
%endif
%endif
;-------
IntCall:mov ah,0
	cwde				;(and eax,0ffh)
%ifdef BEOS
	push edi
	push byte 0
	push edi
	push edx
	push ecx
	push ebx
	push dword be_ret
	int 25h
be_ret:	pop ebx
	pop ebx
	pop ecx
	pop edx
	pop edi
	pop edi
	mov [errno],eax
	and dword [errno],7Fh
	or eax,eax			;set flags also
	pop edi
%else
%ifdef LINUX
	int 80h
%else
	push esi
	push edi
	push edx
	push ecx
	push ebx
	push eax
	call 7:0
	pop ebx
	pop ebx
	pop ecx
	pop edx
	pop edi
	pop esi
	jc err
%endif
	neg eax
err:	mov [errno],eax
	neg eax				;set flags also
%endif
	ret
;-------
%ifdef BEOS
IntRdWr:push edx			;used for Read & Write
	push ecx
	push ebx
	push dword be_ret2
	int 25h
be_ret2:pop ebx
	pop ebx
	pop ecx
	pop edx
	mov [errno],eax
	pop eax
	lea eax,[eax+4]			;add eax,4
	push eax
	mov eax,[errno]
	and dword [errno],7Fh
	or eax,eax			;set flags
	ret
%endif
;--------------------------------------------------------------------------
%ifdef BEOS
KeyCtrlKP:ret
%else
Fork:	mov al,2
	jmp short IntCall
;-------
Pipe:	mov al,42
	jmp short IntCall
;-------
Utime:
%ifdef LINUX
	mov al,30
%else
	mov al,138
%endif
	jmp short IntCall
;-------
Dup2:
%ifdef LINUX
	mov al,63
%else
	mov al,90
%endif
	jmp short IntCall
;-------
Execve:
%ifdef LINUX
	mov al,11
%else
	mov al,59
%endif
	jmp short IntCall
;-------
Wait4:
%ifdef LINUX
	mov al,114			;set ecx to buffer!
%else
	mov al,7
%endif
	xor ebx,ebx
	dec ebx
	xor edx,edx
	xor esi,esi
	xor edi,edi			;at least for FreeBSD
	jmp short IntCall
;-------
Unlink:	mov al,10
	jmp short IntCall
;-----------------------------------------------------------------------
%define CAPTURE_STDERR
;
; pipe buffer through child process
;
KeyCtrlKP:mov esi,extext		;WS, Emacs's, Pico's access to sed|ex pipe
	call InputStringWithMessage
	pushf
	call RestKursPos
	popf
	jc ExExEx
KeyVICmdtemp:mov byte [numeriere],1
	pusha
	mov esi,blockpath
	mov eax,[tempfile]
	mov dword [esi],eax
	mov esi,sot
	mov [blockende],ebp
	call SaveBl2
	popa
	jnc SaveOK
ExExEx:	ret				;cannot save buffer to tmp file...
SaveOK:					;...this is handled inside SaveBlock, so just return
%ifdef CAPTURE_STDERR
	mov ebx,sedpipeC0
	call Pipe
	js OSerrJmp0
%endif
	mov ebx,sedpipeB0
	call Pipe
	js OSerrJmp0
%ifdef FREEBSD
	mov [sedpipeB0],eax		;this seems to work, although...
	inc eax				;...I don't know why....?
	inc eax				;Someone can explain?
	mov [sedpipeB1],eax
%endif
	call Fork
	js OSerrJmp0
	or eax,eax
	jz near ForkChild
;-------
%ifdef USE_EX_NO_SED
;
; This is the UNDEFAULT way using "ex -s" (silent ex).
; We save the buffer into a tempfile and
; WRITE the COMMAND (followed by 'wq' command)
; into a pipe to ex's STDIN.
; Then we truncate the current edit buffer
; and read the tempfile just like a WS block.
;
; "Anything you want, you got it"
; "Anything you need, you got it"
; "Anything at all, you got it, Baby ...." (Jeff Lynne/Roy Orbison/Tom Petty 1989)
;
	mov ebx,[ebx]			;i.e. sedpipeB0
	call CloseFile			;unused pipe direction READ
OSerrJmp0:js OSerrJmp1
	mov ebx,[sedpipeB1]
;-------
	xor edx,edx
	mov ecx,optbuffer
	mov esi,ecx
	cld
	dec edx
Bufloop:inc edx
	lodsb
	or al,al
	jnz Bufloop			;count ex cmd length
	call WriteFile			;write to ex program
;-------
	mov ecx,wq_suffix
	mov edx,wq_suffix_len
	call WriteFile			;write "wq" string to ex program
	call CloseFile
OSerrJmp1:js OSerrJmp
	mov ecx,optbuffer		;for return status
	push edi
	call Wait4
	pop edi
	js OSerrJmp
	mov ecx,[ecx]			;getting return status
	shr ecx,8
	and ecx,0xFF
	mov [errno],ecx
	jnz OSerrJmp
;-------
	mov ebx,tempfile
	push ebp
	push edi
	mov edi,sot			;truncate old file
	mov ebp,sot
	mov al,0ah
	xchg byte [ebp],al		;EOF marker
	push eax
	call ReadBlock2
	pop eax
	pop edi
	jc preEx_Ex
	pop ebx				;dummy (not restore ebp)
;-------
	mov ebx,tempfile
	call Unlink
	jns ChkCursPos			;if no Error
;-------
OSerrJmp:call ChkCursPos		;Error in Wait,Close,...
	jmp near OSerror		;TODO: unlink helper file if exists
preEx_Ex:pop ebp			;Error in ReadBlock
	mov byte[sot],al		;restore pre op values if Read Error
	;continue ChkCursPos
;
%else	;------------------ DEFAULT WAY -------------------
;
; This is the DEFAULT way using stream editor "sed -e".
; The default way is quite opposite: we save the buffer
; into a tempfile, then call sed with the operational command
; given on sed's command line and at last we READ the DATA
; from sed's output via a STDIN READ pipe.
;
	push ebx
	mov ebx,[ebx+4]			;i.e. sedpipeB1
	call CloseFile			;unused pipe direction
%ifdef CAPTURE_STDERR
	mov ebx,[sedpipeC1]
	call CloseFile			;unused pipe direction
%endif
OSerrJmp0:js OSerrJmp1
	pop ebx
	mov esi,[maxlen]
	add esi,sot-SEDBLOCK
	mov ebx,[ebx]			;pipe read channel
	mov ecx,sot-SEDBLOCK
	push ebp
	mov ebp,sot
	mov dword[EmaKiSize],0		;forbid any paste
ReadPipe:mov edx,SEDBLOCK
	add ecx,edx
	call ReadFile
	add ebp,eax			;compute eof
	cmp ebp,esi
	jae ReadOK			;sorry, truncating. FIXME: add a message
	cmp eax,edx
	jz ReadPipe
	cmp ebp,sot			;if nothing comes back, keep buffer as is
	jnz ReadOK
	pop edx
	mov ebp,edx
	push esi			;keep stack balanced
ReadOK:	mov byte [ebp],NEWLINE		;EOF marker
	pop esi				;dummy
	call CloseFile
OSerrJmp1:js OSerrJmp
%ifdef CAPTURE_STDERR
	mov ebx,[sedpipeC0]
%ifdef BEEP_IN_VI
	mov ecx,buffercopy
	xor edx,edx
	inc edx
	call ReadFile
	or eax,eax
	je NoStdErrMsg
	call VIBeepForD
%endif
NoStdErrMsg:call CloseFile		;stderr pipe
%endif
	mov byte [changed],CHANGED	;assumption only
	mov ecx,optbuffer		;for return status
	push edi
	call Wait4
	pop edi
	js OSerrJmp
	mov ebx,tempfile
	call Unlink
	js OSerrJmp
	mov ecx,[ecx]			;getting return status
	shr ecx,8
	and ecx,0xFF
	mov [errno],ecx
	jecxz ChkCursPos
OSerrJmp:call ChkCursPos		;TODO: unlink helper file if exists
	jmp near OSerror
%endif
;----------------------------------------------------------------------
ForkChild:mov ebx,[ebx]			;i.e. sedpipeB0
	xor ecx,ecx
%ifdef USE_EX_NO_SED
	call Dup2			;capturing STDIN
	js FCError
	mov ebx,[sedpipeB1]
	call CloseFile			;unused pipe direction STDOUT
	js FCError
	mov ebx,expath
	mov ecx,exargs
%else
	call CloseFile			;unused pipe direction STDIN
	js FCError
	mov ebx,[sedpipeB1]
	inc ecx
	call Dup2			;capturing STDOUT
	js FCError
%ifdef CAPTURE_STDERR
	mov ebx,[sedpipeC0]
	call CloseFile			;unused pipe direction
	inc ecx
	mov ebx,[sedpipeC1]
	call Dup2			;capturing STDERR
	js FCError
%endif
	mov ebx,sedpath
	mov ecx,sedargs
%endif
	xor edx,edx			;no env
	call Execve
	push byte 103
	pop ebx				;set error
FCex:	jmp near Exit2			;in case of error
FCError:mov ebx,[errno]
	jmp short FCex
%endif
;----------------------------------------------------------------------
;
; care about cursor pos
;
ChkCursPos:cmp edi,ebp			;never let run cursor outside buffer
	jbe CCP
	mov edi,ebp
CCP:	ret
;----------------------------------------------------------------------
;
; copy file to a real backup file (for sym linked files only)
;
; expecting ebx==filepath
;	    ecx==bakpath
;
%ifndef BEOS
CopyToBackup:pusha
	push ecx			;backup file path later needed for Utime call
	push ecx
	call OpenFile0
	xchg esi,eax			;save handle to copy of original file
	mov ecx,O_WRONLY_CREAT_TRUNC
	mov edx,[perms]
	pop ebx				;handle to backupfile
	call OpenFile
	xchg ebx,eax
%ifdef LINUX
	mov ecx,[giduid]
	mov edx,ecx
	shr edx,16
	and ecx,0xffff
	call ChownFile
%else
%ifndef BEOS
	mov edx,[gid]
	mov ecx,[uid]
	call ChownFile
%endif
%endif
;-------
	xor edi,edi			;init eof indicator
copylop:push ebx
	mov ebx,esi			;saved orig file handle
	mov ecx,screenbuffer		;used as copy buffer
	mov edx,4096
	call ReadFile
	pop ebx				;backup file handle
	;js...
	cmp eax,edx
	jz notready
	inc edi				;eof found
notready:mov ecx,screenbuffer
	mov edx,eax			;write read count of byte
	call WriteFile
	;js...
	or edi,edi			;eof ?
	jz copylop
;-------
	call CloseFile			;ready: close backup file
	mov ebx,esi
	call CloseFile			;close original file
	pop ebx				;original file path
	mov ecx,accesstime		;i.e. a data structure of 2* 32 bit
	call Utime			;set change time
	popa
	ret
%endif
;----------------------------------------------------------------------
; recursive descent parser for very SIMPLE math calc within the text:
;              1234+56*78=
; place cursor ^<--here and press   ^KN  ^QC  ^XN  #   ^K
;			(for one of WS   PI   EM   VI  NE modes)
; 		this should insert 5602 into text.
;
; Use 0, 0.000001 ,... up to 999999999999.999999 and operators +-*/ .
; CAUTION: +- Signs are not yet implemented. Do not expect high precision.
;
KeyCtrlKN:
%ifdef USE_MATH
	cld				;preserve ebp,edi!
	finit
	mov esi,edi
	mov [stackptr],esp
	xor eax,eax
	mov dword[level],eax
	mov al,'+'
	call RecursPars
	fstsw [x87]
	and byte[x87],1Fh		;any exception flags?
GErr:	jnz near GeneralErr
	dec dword[level]
	jnz GErr
NoAllgFehl:dec esi
	mov edi,esi			;up to here we have read
	lodsb
	cmp al,'='
	jz EquChar
	mov al,'='
	call OutChar
	dec edi
EquChar:inc edi
	fld qword[factor]
	fmulp st1
	fbstp [x87]
	push byte 12			;12 digits
	pop ecx				
	xor dh,dh			;flag for suppressing leading 0
	lea esi,[x87+9]			;9 BCD data byte and sign
	std
	lodsb
	or al,al
	jns plus
	cmp al,0x80
	jnz GErr
	mov al,'-'
	call OutChar
plus:	call OutHlp
	or dh,dh
	jnz dec_dig
	mov al,'0'
	call OutChar
;-------
dec_dig:mov esi,x87
	xor ebx,ebx
	push byte 3
	pop ecx
	cld
dlop:	lodsb
	mov dl,al
	and al,0fh
	jnz nonull
	inc ebx
	mov al,dl
	shr al,4
	jnz nonull
	inc ebx
isnul2:	loop dlop
	ret				;no decimal digits: ready
;-------
nonull:	push byte 6			;6 decimal digits
	pop ecx
	sub ecx,ebx
	mov dh,0xff			;now do not suppress 0
	mov al,'.'
	call OutChar
	lea esi,[x87+2]			;decimal digits pos
OutHlp:	std
	lodsb
	mov dl,al
	shr al,4
	call OutNumber
	dec ecx
	jecxz xret
	mov al,dl
	and al,0fh
	call OutNumber
	loop OutHlp
xret:	ret
;-------
RecursPars:push eax			;op code
RecConti:lodsb
	cmp al,0ah			;EOL?
	jz RecReturn
	cmp al,'='			;end of task?
	jz RecReturn
	cmp al,'!'			;white space?
	jb RecConti
	cmp al,'0'
	jb noNumber
	cmp al,'9'
	ja isErr
	call Number
	mov al,[esp]			;our opcode
	cmp al,'+'
	jz RecConti
	cmp al,'-'
	jz RecConti
RecReturn:pop eax
	ret
;-------
noNumber:lea esp,[esp-16]
	fstp dword [esp]
	call RecursPars
	fld dword [esp]
	lea esp,[esp+16]
	dec dword[level]
	cmp al,'*'
	jnz noMul
	FMULP st1
	jmp short noMul2
noMul:	cmp al,'/'
	jnz noDiv
	FDIVRP st1
noMul2:	jmp short RecConti
noDiv:	cmp al,'-'
	jnz noSub
	FSUBRP st1
	jmp short RecReturn
noSub:	cmp al,'+'
	jnz isErr
	FADDP st1
noSub2:	jmp short RecReturn
;-------
Number:	FLDZ
	FBSTP [x87]			;init buffer
	push byte 12			;read 13 chars  (up to 12 digits)
	pop ecx
num_ctr:call DigitHlp
	jc int_end
	loop num_ctr
isErr:	jmp GeneralErr			;if >12 digits
int_end:push edi			;****
	neg ecx
	lea ecx,[ecx+13]
	push ecx			;stor # of integer digits
	inc dword[level]
	dec esi				;num_ctr loop has read 1 too much
	mov edx,esi			;position we have read so far
	cmp al,('.'-'0')
	jnz integers
	inc esi
	lea edi,[x87+2]			;start of decimal places
	mov cl,3			;6/2 decimal places
dec_ctr:cld
	call DigitHlp
	jc decend
	shl al,4
	mov bl,al
	mov [edi],al			;important if abort at digit 1,3,5
	call DigitHlp
	jc decend	
	add al,bl
	std
	stosb
	loop dec_ctr			;if >6 decimal places it will run into error later
	inc esi
decend:	dec esi
	xchg edx,esi			;edx where later to continue scanning
;-------
integers:pop ecx			;ecx # of integer digits
	dec esi				;esi where integer places are
	lea edi,[x87+3]			;start of integer part
intloop:std
	lodsb
	sub al,'0'
	mov bl,al
	dec ecx
	jecxz h2
	lodsb
	sub al,'0'
	shl al,4
	add al,bl
h2:	cld
	stosb
	jecxz h3
	loop intloop
h3:	cld
	FBLD [x87]
	FLD qword [factor]
	fdivp st1
	mov esi,edx			;pointer for continued reading
	pop edi				;****
	ret
;-------
DigitHlp:lodsb
	sub al,'0'
	jb dret
	cmp al,10
	cmc
dret:	ret				;return: al=value / cy if error
;--------
OutNumber:cmp al,dh			;flag set?
	jz OCret
	add al,'0'
	mov dh,0xff			;set flag
OutChar:push esi
	push edx
	push ecx
	call NormChar
	pop ecx
	pop edx
	pop esi
	inc ebx
%endif
OCret:	ret
;----------------------------------------------------------------------
;
; CONSTANT DATA AREA
;
factor		dq 1000000.0
%ifndef BEOS
tempfile 	db 'e3$$',0		;tempfile (FIXME: use PID for name)

%ifdef USE_EX_NO_SED
exargs		dd expath
		dd minus_s
		dd tempfile
		dd 0
expath		db EX_PATH,0
minus_s		db '-s',0
wq_suffix	db 0ah,'wq',0ah
wq_suffix_len	equ $-wq_suffix
%else
sedargs		dd sedpath		;this way default
		dd minus_e
		dd optbuffer
		dd tempfile
		dd 0
sedpath		db SEDPATH,0
minus_e		db '-e',0
%endif
%endif
;
optiontext	db 'OPT? C/B',0
filename	db 'FILENAME:',0
block		db '   NAME:',0
saveas		db 'SAVE AS:',0
filesave	db '   SAVE:',0
asksave		db 'SAVE? Ynl',0
asksave2	db 'SAVE? Yn',0
askreplace1	db 'REPLACE:',0
askreplace2	db 'RE WITH:',0
asklineno	db 'GO LINE:',0
askfind		db ' SEARCH:',0
asknumber	db '^Q OCTAL:',0
extext		db 'mode EX:',0
modetxt		db 'SET MODE',0

%define DoNo 10
ScanTable: 	;another xlat table containing offsets in jumptab1 table
	db DoNo	;		esc[0~
	db 0	;keyHome	esc[1~
	db 8	;keyIns		esc[2~
	db 9	;keyDel		esc[3~
	db 5	;keyEnd		esc[4~
	db 2	;keyPgUp	esc[5~
	db 7	;KeyPDn		esc[6~
	db 0	;keyHome	esc[7~
	db 5	;keyEnd		esc[8~
		;---------------------
	db 1	;keyUp		esc[A
	db 6	;keyDown	esc[B
	db 4	;keyRight	esc[C
	db 3	;keyLeft	esc[D
	db DoNo	;		esc[E
	db 5	;keyEnd		esc[F
	db 7	;keyPgDn	esc[G
	db 0	;keyHome	esc[H
%ifndef LINUX
	db 2	;keyPUp		esc[I
	db DoNo	;		esc[J
	db DoNo	;		esc[K
	db 8	;keyIns		esc[L
%endif
STsize equ ($-ScanTable)
;----------------------------------------------------------------------
EmaAltTable: 	;another xlat table containing offsets in jumptab1 table
	db 12h	;'B'
	db DoNo	;'C'
	db DoNo	;'D'
	db DoNo	;'E'
	db 13h	;'F'
	db 11h	;'G'
	db 3Dh	;'H'	Help!
	db DoNo	;'I'
	db DoNo	;'J'
	db DoNo	;'K'
	db DoNo	;'L'
	db DoNo	;'M'
	db DoNo	;'N'
	db DoNo	;'O'
	db DoNo	;'P'
	db DoNo	;'Q'
	db DoNo	;'R'
	db DoNo	;'S'
	db DoNo	;'T'
	db DoNo	;'U'
	db 2 	;'V'
	db 27h	;'W'
	db 3Eh	;'X'
ATsize equ ($-EmaAltTable)
;----------------------------------------------------------------------
%define Beep 0x4E
VIcmdTable:db Beep;0
	db Beep	;1
	db 2	;^B PageUp
	db Beep	;3
	db 36h	;^D half PageUp
	db Beep	;5
	db 7	;^F PageDn
	db Beep	;7
	db 3	;^H KeyLeft
	db Beep	;9
	db 6	;^J KeyDown
	db Beep	;11
	db Beep	;12
	db 6	;^M KeyDown
	db Beep	;^N
	db Beep	;^O
	db Beep	;^P
	db Beep	;^Q
	db Beep	;^R
	db Beep	;^S
	db Beep	;^T
	db 37h 	;^U
	db Beep	;22
	db Beep	;23
	db Beep	;24
	db Beep	;25
	db Beep	;26
	db Beep	;27
	db Beep	;28
	db Beep	;29
	db Beep	;30
	db Beep	;31
	db 4	;' ' KeyRight
	db Beep	;33
	db Beep	;34
	db 4fH 	;35  Numerics
	db 5	;'$' KeyEnd
	db Beep	;37
	db Beep	;38
	db Beep	;39
	db Beep	;40
	db Beep	;41
	db Beep	;42
	db 6	;'+' KeyDown
	db Beep	;44
	db 1	;'-' KeyUp
	db Beep	;46
	db 39h	;'/' Search
	db 0	;'0' KeyHome
	db 44h	;'1' 1G BOF
	db Beep	;'2'
	db Beep	;'3'
	db Beep	;'4'
	db Beep	;'5'
	db Beep	;'6'
	db Beep	;'7'
	db Beep	;'8'
	db Beep	;'9'
	db 2Ch	;':' ex mode
	db 3Eh	;';' e3 special command: QUICK leave vi mode :-)  press  e3ws, e3em, e3pi, e3ne
	db Beep	;'<'
	db Beep	;'='
	db Beep	;'>'
	db 3Ah	;'?' search backw
	db Beep	;'@'
	db 2Eh	;'A'
	db 12h	;'B' left word
	db 4Bh	;'C' Change rest of line 
	db 4Ch	;'D' Delete rest of line (not unlike ^QY in WStar)
	db Beep	;'E'
	db Beep	;'F'
	db 0Fh	;'G' EOF
	db 30h	;'H' First LIne
	db 33h	;'I' switch to insert mode
	db 4Dh	;'J' Join lines
	db Beep	;'K'
	db 31h	;'L' Last Line
	db Beep	;'M'
	db Beep	;'N'
	db 32h	;'O' Open Line
	db 3Ch	;'P' Paste
	db Beep	;'Q'
	db 34h	;'R' overwrite
	db 40h	;'S' kill +insmode
	db Beep	;'T'
	db Beep	;'U'
	db Beep	;'V'
	db 13h	;'W' next word
	db 3fh	;'X' del left
	db Beep	;'Y'
	db 41h	;'Z'
	db Beep	;'['
	db Beep	;'\'
	db Beep	;']'
	db 38h	;'^' KeyFirstChar
	db Beep	;'_'
	db Beep	;'`'
	db 2Dh	;'a' append (KeyLeft plus insert mode)
	db 12h	;'b' left word
	db Beep	;'c'
	db 35h	;'d' delete
	db 43h	;'e'
	db Beep	;'f'
	db Beep	;'g'
	db 3	;'h' KeyLeft
	db 2Bh	;'i' switch to insert mode
	db 6	;'j' KeyDown
	db 1	;'k' KeyUp
	db 4	;'l' KeyRight
	db 45h	;'m' set the one and only marker 'a'
	db Beep	;'n'
	db 2Fh	;'o' Open Line
	db 3Bh	;'p' Paste
	db Beep	;'q'
	db 4Ah	;'r' repl one char
	db Beep	;'s'
	db Beep	;'t'
	db Beep	;'u'
	db Beep	;'v'
	db 13h	;'w' next word
	db 9	;'x' KeyDel
	db 46h	;'y'
	db 42h	;'z' center line z.
VIsize equ ($-VIcmdTable)
;----------------------------------------------------------------------
Ktable	db DoNo	;^K@	xlatb table for making pseudo-scancode
	db DoNo	;^ka
	db 24h	;^kb	24h for example points to KeyCtrlKB function offset
	db 15h	;^kc
	db 0dh	;^kd
	db DoNo	;^ke	DoNo means SimpleRet i.e. 'do nothing'
	db DoNo	;^kf
	db DoNo	;^kg
	db 1dh	;^kh
	db DoNo	;^ki
	db DoNo	;^kj
	db 14h	;^kk
	db DoNo	;^kl
	db 3eh	;^km	Set Mode
	db 4FH	;^kn
	db DoNo	;^ko
	db 48h 	;^kp    Pipe thru sed
	db 0bh	;^kq
	db 10h	;^kr
	db 0ch	;^ks
	db DoNo	;^kt
	db DoNo	;^ku
	db 23h	;^kv
	db 25h	;^kw
	db 16h	;^kx
	db 19h	;^ky
	db DoNo	;^kz
Qtable	db DoNo	;^q@	ditto for ^Q menu
	db 1ah	;^qa
	db 20h	;^qb
	db 0fh	;^qc
	db 05h	;^qd
	db 1eh	;^qe
	db 1bh	;^qf
	db DoNo	;^qg
	db 17h	;^qh, ^qDEL
	db 11h	;^qi
	db DoNo	;^qj
	db 21h	;^qk
	db DoNo	;^ql
	db DoNo	;^qm
	db DoNo	;^qn
	db DoNo	;^qo
	db 18h	;^qp
	db DoNo	;^qq
	db 0eh	;^qr
	db 00h	;^qs
	db DoNo	;^qt
	db DoNo	;^qu
	db 1ch	;^qv
	db 12h	;^qw
	db 1fh	;^qx
	db 22h	;^qy
	db 13h	;^qz
;----------------------------------------------------------------------
Xtable	db DoNo	;^x^@
	db DoNo	;^x^a
	db DoNo	;^x^b
	db 0bh	;^x^c	WS: ^KQ
	db DoNo	;^x^d
	db DoNo	;^x^e
	db 47h	;^x^f
	db DoNo	;^x^g
	db 3Dh	;^x^h   		i.e.  "HELP!" (The Beatles, 1965)  ;-)
	db 10h	;^x i	WS: ^KR
	db DoNo	;^x^j
	db DoNo	;^x^k
	db DoNo	;^x^l
	db DoNo	;^x^m
	db 4fh	;^x^n   numerics
	db DoNo	;^x^o
	db 48h	;^x^p   special sed pipe
	db DoNo	;^x^q
	db DoNo	;^x^r
	db 0ch	;^x^s	WS: ^KS
 	db DoNo	;^x^t
	db DoNo	;^x^u
	db DoNo	;^x^v
	db 29h	;^x^w	write to
	db 26h	;^x^x	xchg mark/point
	db DoNo	;^x^y
	db DoNo	;^x^z
;----------------------------------------------------------------------
PicoJtable db DoNo ;^j@	Junk ops for PI mode
	db DoNo	;^ja
	db DoNo	;^jb
	db DoNo	;^jc
	db DoNo	;^jd
	db DoNo	;^je
	db DoNo	;^jf
	db DoNo	;^jg
	db 17h 	;^jh Junk to line Home
	db DoNo	;^ji
	db DoNo	;^jj
	db DoNo	;^jk
	db 22h	;^jl Junk Line rest
	db DoNo	;^jm
	db DoNo	;^jn
	db DoNo	;^jo
	db 48h	;^jp  special sed pipe
	db DoNo	;^jq
	db DoNo	;^jr
	db DoNo	;^js
	db 49h	;^jt repeat last search&replace
	db DoNo	;^ju
	db DoNo	;^jv
	db 2Ah	;^jw Junk Word
	db DoNo	;^jx
	db DoNo	;^jy
	db DoNo	;^jz
PicoQtable db DoNo ;^q@	Quick motions for PI mode: wordstar counterparts on different keys
	db DoNo	;^qa
	db 1Fh	;^qb	Bottom of window
	db 4FH	;^qc	Calc numerics
	db DoNo	;^qd
	db 0fh	;^qe	End of file
	db 1Ch	;^qf	last Find
	db DoNo	;^qg
	db DoNo	;^qh
	db DoNo	;^qi
	db DoNo	;^qj
	db DoNo	;^qk
	db 11h	;^ql	Line number #
	db 3eh	;^qm	set mode
	db 13h	;^qn	Next word
	db DoNo	;^qo
	db 12h	;^qp	Previous word
	db DoNo	;^qq
	db DoNo	;^qr
	db 0Eh	;^qs	Start of file
	db 1Eh	;^qt	Top of window
	db DoNo	;^qu
	db DoNo	;^qv
	db DoNo	;^qw
	db DoNo	;^qx
	db DoNo	;^qy
	db DoNo	;^qz
;----------------------------------------------------------------------
esize equ 2	;(byte per entry)
jumptab1:	;Storing 16 bit offsets is valid only for code less size 64 kbyte...
		;  ... but in assembler that should never be a problem   ;)
	dw KeyHome	-_start	;0
	dw KeyUp	-_start	;1
	dw KeyPgUp	-_start	;2
	dw KeyLeft	-_start	;3
	dw KeyRight	-_start	;4
	dw KeyEnd	-_start	;5
	dw KeyDown	-_start	;6
	dw KeyPgDn	-_start	;7
	dw KeyIns	-_start	;8
	dw KeyDel	-_start	;9	0..9 are Cursor pad keys
;------------------------
	dw SimpleRet	-_start	;10 	DO_NOTHING == DoNo
	dw KeyCtrlKQ	-_start	;0bh EMA ^X^C
	dw KeyCtrlKS	-_start	;0ch EMA ^X^S
	dw KeyCtrlKD	-_start	;0dh EMA ^X^F
	dw KeyCtrlQR	-_start	;0eh EMA Alt<
	dw KeyCtrlQC	-_start	;0fh EMA Alt>
	dw KeyCtrlKR	-_start	;10h EMA ^XI
	dw KeyCtrlQI	-_start	;11h EMA Alt-G
	dw KeyCtrlQW	-_start	;12h EMA Alt-B
	dw KeyCtrlQZ	-_start	;13h EMA Alt-F
;--------------------------
;up to here this functions are considered common for all (exc vi)
;Of course some use different key names.
;
;now follows special stuff for each editor emulation:
;-------WS and Pico--------
	dw KeyCtrlKK	-_start	;14h
	dw KeyCtrlKC	-_start	;15h
	dw KeyCtrlKX	-_start	;16h
	dw KeyCtrlQDel	-_start	;17h
	dw KeyCtrlQP	-_start	;18h
	dw KeyCtrlKY	-_start	;19h
	dw KeyCtrlQA	-_start	;1ah
	dw KeyCtrlQF	-_start	;1bh
	dw KeyCtrlQV	-_start	;1ch
	dw KeyCtrlKH	-_start	;1dh
	dw KeyCtrlQE	-_start	;1eh
	dw KeyCtrlQX	-_start	;1fh
	dw KeyCtrlQB	-_start	;20h
	dw KeyCtrlQK	-_start	;21h
	dw KeyCtrlQY	-_start	;22h
	dw KeyCtrlKV	-_start	;23h
	dw KeyCtrlKB	-_start	;24h
	dw KeyCtrlKW	-_start	;25h
;-------EM--------
	dw KeyCtrlXX	-_start	;26h
	dw KeyEmaAltW	-_start	;27h
	dw KeyEmaAltPer	-_start ;28h Alt-%
	dw KeyEmaCtrlXW	-_start	;29h
;-------PI--------
	dw KeyCtrlT	-_start	;2Ah
;-------VI--------
	dw KeyVIcmdi	-_start	;2Bh
	dw KeyVIex	-_start	;2Ch
	dw KeyVIcmda	-_start ;2Dh
	dw KeyVICmdA	-_start	;2Eh
	dw KeyVICmdo	-_start	;2Fh
	dw KeyCtrlQE	-_start	;30h
	dw KeyCtrlQX	-_start	;31h
	dw KeyVICmdO	-_start	;32h
	dw KeyVICmdI	-_start	;33h
	dw KeyVICmdR	-_start ;34h
	dw KeyVICmdd	-_start ;35h
	dw KeyHalfPgDn	-_start ;36h
	dw KeyHalfPgUp	-_start ;37h
	dw KeyVI1Char	-_start ;38h
	dw KeyVIfsearch	-_start ;39h
	dw KeyVIbsearch	-_start ;3Ah
	dw KeyVICmdp	-_start ;3Bh
	dw KeyVICmdP	-_start ;3Ch
;-------later added------
	dw KeyHelp	-_start ;3Dh	general
	dw KeyEditMode	-_start ;3Eh	general
	dw KeyDell	-_start ;3Fh	vi
	dw KeyVICmdS	-_start ;40h	vi	
	dw KeyVICmdZ	-_start ;41h	vi
	dw KeyVICmdz	-_start ;42h	vi
	dw KeyVIcmde	-_start ;43h	vi
	dw KeyVIcmd1	-_start ;44h	vi
	dw KeyVICmdm	-_start ;45h	vi
	dw KeyVICmdy	-_start ;46h	vi
	dw KeyEmaCtrlXF	-_start ;47h	emacs (extended ^KD from WS)
	dw KeyCtrlKP	-_start ;48h    use sed-pipe in WS,Emacs,Pico
	dw KeyPiCtrlJT	-_start ;49h	Pico
	dw KeyVICmdr	-_start ;4Ah	vi
	dw KeyVICmdC	-_start ;4Bh	vi
	dw KeyVICmdD	-_start ;4Ch	vi
	dw KeyVICmdJ	-_start ;4Dh	vi
%ifdef BEEP_IN_VI
	dw VIBeepForD	-_start ;4Eh	vi
%else
	dw SimpleRet	-_start ;4Eh	vi
%endif
	dw KeyCtrlKN	-_start ;4Fh
jumps1 equ ($-jumptab1) / esize
;--- 32 more for WS--------
	dw SimpleRet	-_start	;^Space
	dw KeyCtrlQW	-_start	;^a
	dw SimpleRet	-_start	;^b
	dw KeyPgDn	-_start	;^c
	dw KeyRight	-_start ;^d
	dw KeyUp	-_start	;^e
	dw KeyCtrlQZ	-_start	;^f
	dw KeyDel	-_start	;^g 7
	dw KeyDell	-_start	;^h 8   DEL (7fh is translated)
	dw NormChar	-_start	;^i 9	(TAB)
;;;	dw KeyRetNoInd	-_start	;^j
	dw KeyHelp	-_start	
	dw CtrlKMenu	-_start	;^k b
	dw KeyCtrlL	-_start	;^l c
	dw KeyRet	-_start	;^m 0dh
	dw SimpleRet	-_start	;^n e
	dw SimpleRet	-_start	;^o f
	dw KeyHelp	-_start	;^p 10	Help!
	dw CtrlQMenu	-_start	;^q 11
	dw KeyPgUp	-_start	;^r 12
	dw KeyLeft	-_start	;^s 13
	dw KeyCtrlT	-_start	;^t 14
	dw SimpleRet	-_start	;^u 15 (abort in Input)
	dw KeyIns	-_start	;^v 16
	dw KeyScrollUp	-_start	;^w 17
	dw KeyDown	-_start	;^x 18
	dw KeyCtrlY	-_start	;^y 19
	dw KeyScrollDn	-_start ;1a
	dw SimpleRet	-_start ;1b
	dw SimpleRet	-_start	;1c
	dw SimpleRet	-_start	;1d
	dw SimpleRet	-_start	;1e
	dw SimpleRet	-_start	;1f
;--- 32 more for EM--------
	dw KeyEmaMark	-_start	;^Space
	dw KeyHome	-_start	;^a
	dw KeyLeft	-_start	;^b
	dw SimpleRet	-_start	;^c (not planned)
	dw KeyDel	-_start ;^d
	dw KeyEnd	-_start	;^e
	dw KeyRight	-_start	;^f
	dw SimpleRet	-_start	;^g (abort in Input)
	dw KeyDell	-_start	;^h
	dw NormChar	-_start	;^i (TAB)
	dw KeyRet	-_start	;^j
	dw KeyEmaCtrlK	-_start	;^k
	dw KeyEmaCtrlL	-_start	;^l
	dw KeyRetNoInd	-_start	;^m 0dh
	dw KeyDown	-_start	;^n
	dw KeyEmaCtrlO	-_start	;^o
	dw KeyUp	-_start	;^p
	dw KeyEmaCtrlQ	-_start	;^q
	dw KeyEmaCtrlR	-_start	;^r
	dw KeyEmaCtrlS	-_start	;^s
	dw KeyEmaCtrlT	-_start	;^t
	dw SimpleRet	-_start	;^u (not yet planned)
	dw KeyPgDn	-_start	;^v
	dw KeyEmaCtrlW	-_start	;^w
	dw CtrlXMenu	-_start	;^x
	dw KeyEmaCtrlY	-_start	;^y
	dw SimpleRet	-_start ;1a
	dw SimpleRet	-_start ;1b
	dw SimpleRet	-_start	;1c
	dw SimpleRet	-_start	;1d
	dw SimpleRet	-_start	;1e
	dw SimpleRet	-_start	;1f
;--- 32 more for PI------
	dw KeyEmaMark	-_start	;^Space	a redundant marker because ^^ is ugly on some kbds
	dw KeyHome	-_start	;^a 1
	dw KeyLeft	-_start	;^b 2
	dw SimpleRet	-_start	;^c 3
	dw KeyDel	-_start ;^d 4
	dw KeyEnd	-_start	;^e 5
	dw KeyRight	-_start	;^f 6
	dw KeyHelp	-_start	;^g 7
	dw KeyDell	-_start	;^h 8   DEL
	dw NormChar	-_start	;^i 9	(TAB)
	dw PicoJMenu	-_start	;^j a
	dw KeyEmaCtrlW	-_start	;^k b
	dw KeyEmaMark	-_start	;^l c	a redundant marker because ^^ is ugly on some kbds
	dw KeyRet	-_start	;^m d
	dw KeyDown	-_start	;^n e
	dw KeyCtrlKS	-_start	;^o f	SAVE
	dw KeyUp	-_start	;^p 10
	dw PicoQMenu	-_start	;^q 11
	dw KeyCtrlKR	-_start	;^r 12
	dw KeyEmaCtrlXW	-_start	;^s 13	SAVE_AS
	dw KeyEmaAltPer	-_start	;^t 14
	dw KeyEmaCtrlY	-_start	;^u 15
	dw KeyPgDn	-_start	;^v 16
	dw KeyEmaCtrlS	-_start	;^w 17
	dw KeyCtrlKQ	-_start	;^x 18
	dw KeyPgUp	-_start	;^y 19
	dw SimpleRet	-_start	;^z 1A
	dw SimpleRet	-_start	;^[ 1B
	dw SimpleRet	-_start	;^\ 1C
	dw SimpleRet	-_start	;^] 1D
	dw KeyEmaMark	-_start	;^^ 1E	see ^L
	dw SimpleRet	-_start ;1F
;--- 32 more for NE------
	dw KeyEmaMark	-_start	;^Space toggle selection mode (no shift cursor keys available!)
	dw KeyNedCtrlA	-_start	;^a 1   Mark all
	dw KeyIns	-_start	;^b 2   toggle Ins mode
	dw KeyEmaAltW	-_start	;^c 3	COPY
	dw SimpleRet	-_start ;^d 4
	dw KeyEditMode	-_start	;^e 5   set EDit mode
	dw KeyCtrlQF	-_start	;^f 6   find
	dw KeyCtrlL	-_start	;^g 7   find again
	dw KeyDell	-_start	;^h 8   DEL
	dw NormChar	-_start	;^i 9	TAB
	dw KeyRet	-_start	;^j a   RETURN
	dw KeyCtrlKN	-_start	;^k b   numerics
	dw KeyCtrlQI	-_start	;^l c   LINE #
	dw KeyRet	-_start	;^m d   RETURN
	dw KeyEmaCtrlXF	-_start	;^n e	OPEN another
	dw KeyEmaCtrlXF	-_start	;^o f   OPEN another
	dw SimpleRet	-_start	;^p 10
	dw KeyCtrlKQ	-_start	;^q 11  EXIT
	dw KeyCtrlQA	-_start	;^r 12  REPLACE
	dw KeyCtrlKS	-_start	;^s 13	SAVE
	dw SimpleRet	-_start	;^t 14
	dw SimpleRet	-_start	;^u 15
	dw KeyEmaCtrlY	-_start	;^v 16	PASTE
	dw KeyEmaCtrlXW	-_start	;^w 13	SAVE_AS/WRITE TO
	dw KeyEmaCtrlW	-_start	;^x 18	CUT
	dw SimpleRet	-_start	;^y 19
	dw SimpleRet	-_start	;^z 1A
	dw SimpleRet	-_start	;^[ 1B
	dw SimpleRet	-_start	;^\ 1C
	dw SimpleRet	-_start	;^] 1D
	dw SimpleRet	-_start	;^^ 1E
	dw SimpleRet	-_start ;1F
;----------------------------------------------------------------------
;
BeepChar	db 7
screencolors0	db 27,'[40m',27,'[37m'
bold0		db 27,'[0m'		;reset to b/w
screencolors1	db 27,'[44m',27,'[33m'	;yellow on blue
reversevideoX:
bold1:		db 27,'[1m'		;bold
scolorslen	equ $-screencolors1
boldlen		equ $-bold1		;take care length of bold0 == length of bold1
%ifdef LINUX
		db 27,'[7m'		;good for "linux" terminal on /dev/tty (but not xterm,kvt)
					;again take care length = length of boldX
					;!! important: store directly after bold1 !!
%endif
;-------------------------------------------------------------------------
editmode:db 'p WSp Pip Emp vip NE'
;
helptext:
db "MicroEditor e3 v1.61 / GPL (C) 2000,2001 A.Kleine <kleine@ak.sax.de>",10
db "Enter filename or leave with RETURN.",10,10
helptextsize equ $-helptext					;keep <= 127 byte
;
helpfoot:db 10,10,10,"			-= PRESS ANY KEY =-"	;at least 6 wasted byte ;-)
helpfootsize equ $-helpfoot
;
help_ws:
db "Key bindings in WS mode:",10,10
db "Files:	^KR Insert	^KS Save	^KX Save&Exit	^KQ Abort&Exit",10
%ifdef USE_EX_NO_SED
db "	^KD Save&Load	^KM Set mode	^KP Pipe buffer through 'ex' ",10
%else
db "	^KD Save&Load	^KM Set mode	^KP Pipe buffer through 'sed'",10
%endif
db 10
db "Blocks:	^KB Start	^KK End		^KC Copy	^KY Delete",10
db "	^KV Move	^KW Write",10
db 10
db "Search:	^QF Find	^L  Repeat	^QA Search&Repl.",10
db "Move:	^E  Up		^X  Down	^S  Left	^D  Right",10
db "	^R  Page Up	^C  Page Down	^W  Scroll Up	^Z  Scroll Dn",10
db "Quick-	^QE Wnd Top	^QX Wnd Bottom	^QS Home	^QD End",10
db "-Move:	^QR BOF		^QC EOF		^QB Blk Begin	^QK Blk End",10
db "	^F  Next Word	^A  Prev Word	^QI Line#	^QV Last Find",10
db 10
db "Delete:	^T  Word	^Y  Line	^H  Left	^G  Char",10
db "	^QY Line End	^QDel,^QH Line Begin",10
db "Other:	^KN Numerics",10
help_ws_size equ $-help_ws
;-------------------------
help_pi:
db "Key bindings in PICO mode:",10,10
db "Files:	^XN ExitNoSave	^XY Exit+Save	^XL Save+Load New File",10
db "	^O  Save	^S  Save as	^R  Read",10
db 10
db "Move:	^P  Up		^N  Down	^B  Left	^F  Right",10
db "	^Y  Page up	^V  Page down	^QN Next word	^QP Previous word",10
db "	^A  Home	^E  End		^QS Start	^QE EOF",10
db "	^QT Top screen	^QB Bottom scr	^QL to Line#	^QF last Find",10
db 10
db "Search:	^W  Where is	^T  Search&Repl	^JT Repeat Search & Replace",10
db 10
db "Delete:	^H  Left char	^D  This char	^K  Kill line/region",10
db "	^JW Word	^JL Line end	^JH To Line begin",10
db 10
db "Other:	^U  Unkill	^G  Help	^^,^L,^<SPC> Mark region",10
%ifdef USE_EX_NO_SED
db "	^QM Set Edit Mode		^JP Pipe buffer through 'ex' ",10
%else
db "	^QM Set Edit Mode		^JP Pipe buffer through 'sed'",10
db "	^QC Calculate Numerics"
%endif

help_pi_size equ $-help_pi
;-------------------------
help_em:
db "Key bindings in EMACS mode:",10,10
db "Files:	^X^C Exit	^XI  Insert	^X^S Save	^X^F Load New",10
%ifdef USE_EX_NO_SED
db "	^X^W Write new	^X^H Help	^X^P Pipe buffer through 'ex' ",10
%else
db "	^X^W Write new	^X^H Help	^X^P Pipe buffer through 'sed'",10
%endif
db 10
db "Move:	^P   Up		^N  Down	^B   Left	^F   Right",10
db "	altV Pg up	^V  Pg down	altB Left word	altF Right word",10
db "	^A   Home	^E  End		alt< BOF	alt> EOF",10
db "	altG Go line#	^L  Center Pos",10
db 10
db "Search:	^S Find fwd	^R Find bwd	alt% Search&Replace like WS",10
db 10
db "Buffer:	altW Copy	^Y Yank		^<SPC> Mark	^X^X Xchg Mark/Pt",10
db 10
db "Delete:	^K Line		^W Region	^H Left	Char	^D This Char",10
db 10
db "Other:	^O Open line	^T Xchg char	^I Insrt Tab	^Q Quoted Ins",10
db "	^M NL		^J NL+indent	altX Set edit mode",10
db "	^X^N Calculate Numerics"
help_em_size equ $-help_em
;-------------------------
help_vi:
db "Key bindings in vi mode:",10
db 10
db "<ESC>			enter cmd mode",10
db "h,j,k,l,+,-,<Ret>,<SPC>	move by chars&lines",10
db "^B,^F,^D,^U		move by [half]page",10
db "$,0,^,w,b,e,H,L,z.	more move in line/screen",10
db "/,?			search forward/backward",10
db "ma			set marker a",10
db "x,X,<Del>,D		delete char, to EOL",10
db "dd,S,C,d'a,y'a,yy	kill,substitute,change,delete,yank",10
db "p,P			paste",10
db "A,a,I,i,<Ins>,O,o	enter insert mode",10
db "R,r			enter ovw.mode, ovw.char",10
db "J			join lines",10
db "ZZ			save&quit",10
db ";,#			E3 SPECIAL: set edit mode, calculate",10
db ":w,:wq,:q,:q!,:e	ex mode:save,quit,save_as,edit other",10
db ":0,:$,:<line#>		ex mode:go BOF,EOF,line",10
db ":h			ex mode:help",10
%ifdef USE_EX_NO_SED
db ":<other cmd>		pipe buffer through 'ex' "
%else
db ":<other cmd>		pipe buffer through 'sed'"
%endif
help_vi_size equ $-help_vi
;-------------------------
help_ne:
db "Key bindings in NEDIT mode:",10
db 10
db "Files:		^QN Exit_NoSave	^QY Exit&Save	^QL  Save&Load new file",10
db "		^S  Save	^W  WriteTo=SaveAs	",10
db "		^O,^N Other file		altI Insert file",10
db 10
db "Move:		Cursor keys (as usual)		^L   Go Line #",10
db 10
db "Select:		^<SPACE> begin & extend using cursor keys (like Emacs mode!)",10
db "		^A  select All buffer",10
db "		^X  Cut		^C  Copy 	^V  Paste",10
db 10
db "Delete:		^H,Del<- Left char		Del This character",10
db 10
db "Search:		^F   Find	^R Search and Replace (just like WS mode!)",10
db "		^G   Go and repeat last Find or last Search & Replace",10
db 10
db "Other:		^E   Set editor mode: e3ws/e3em/e3pi/e3vi/e3ne",10
db "		^K   Calculate numerics",10
db "		altH Help key bindings"

help_ne_size equ $-help_ne

%if help_ws_size != help_pi_size || help_ws_size!= help_em_size || help_ws_size!= help_pi_size || help_ws_size!= help_vi_size || help_ws_size!= help_ne_size
%error Helptext
dw help_vi_size
dw help_ws_size
dw help_pi_size
dw help_em_size
dw help_ne_size
%endif
;
;----------------------------------------------------------------------
errmsgs equ $
db "Operation not permitted",10
db "No such file|directory",10
db 10
db "Interrupted sys call",10
db "Input/output",10
db "Device not configured",10
db 10
db 10
db "Bad file descriptor",10
db "No child processes",10
db "Resource temp.unavailable",10
db "Cannot alloc.memory",10
db "Permission denied",10
db 10
db 10
db "Device|resource busy",10
db "File exists",10
db 10
db "No such device",10
db "Not a directory",10
db "Is a directory",10
db "Invalid argument",10
db "Too many open files i.system",10
db "Too many open files",10
db "Inappropriate ioctl f.device",10
db "Text file busy",10
db "File too large",10
db "No space on device",10
db "Illegal seek",10
db "ReadOnly file system",10
db 10
db "Broken pipe",10
db 10
db 10
db 10
db "File name too long",10
db 10
db "Function not implemented",10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db "No data available",10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db "File descriptor in bad state",10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db "Operation not supported",10
db 10
db 10
db 10
db 10
db 10
db 10
db 10
db "Can't exec "
%ifdef USE_EX_NO_SED
db EX_PATH,10
%else
db SEDPATH,10
%endif
db 10
db "No buffer space available",10
;
;
%ifdef CRIPLED_ELF
 filesize      equ     $ - $$
%endif
;-----------------------------------------------------------------------
;section .data				;unused: save byte in ELF header
;bits 32
;-----------------------------------------------------------------------
section .bss
bits 32

%ifdef CRIPLED_ELF
 align 4
 bssstart:
%endif

screenbuffer	resb 62*(160+32)	;estimated 62 lines 160 columns, 32 byte ESC seq (ca.12k)
screenbuffer_end equ $			;If you really have higher screen resolution,
					;...no problem, except some useless redrawing happens.
termios_size	equ 60
termios		resb termios_size
orig		resb termios_size

winsize_size	equ 8
winsize		resb winsize_size

setkplen	equ 10
setkp		resb setkplen		;to store cursor ESC seq like  db 27,'[000;000H'
resb		2			;fill up

revvoff		resd 1

lines		resd 1			;equ 24 or similar i.e. screen lines-1 (1 for statusline)
columns		resd 1			;equ 80 or similar dword (using only LSB)

columne		resd 1			;helper for display of current column
zloffst		resd 1			;helper: chars scrolled out at left border
fileptr		resd 1			;helper for temp storage of current pos in file
kurspos		resd 1			;cursor position set by DispNewScreen()
kurspos2	resd 1			;cursor position set by other functions

tabcnt		resd 1			;internal helper byte in DispNewScreen() only
changed		resd 1			;status byte: (UN)CHANGED
oldQFpos	resd 1
bereitsges	resd 1			;byte used for ^L

blockbegin	resd 1
blockende	resd 1
endeedit	resd 1			;byte controls program exit
old		resd 1			;helper for ^QP
veryold		resd 1			;ditto
linenr		resd 1			;current line
showblock	resd 1			;helper for ^KH
suchlaenge	resd 1			;helper for ^QA,^QF
repllaenge	resd 1
vorwarts	resd 1
grossklein	resd 1			;helper byte for ^QF,^QA

ch2linebeg	resd 1			;helper keeping cursor pos max at EOL (up/dn keys)
numeriere	resd 1			;byte controls re-numeration
read_b		resd 1			;buffer for getchar
isbold		resd 1			;control of bold display of ws-blocks
inverse		resd 1
insstat		resd 1
errno		resd 1			;used similar libc, but not excactly equal

errlen		equ 100
error		resb errlen		;reserved space for string: 'ERROR xxx:tttteeeexxxxtttt',0

maxlen		resd 1
;-------
;
maxfilenamelen	equ 255
filepath	resb maxfilenamelen+1
bakpath		resb maxfilenamelen+1
blockpath	resb maxfilenamelen+1
replacetext	resb maxfilenamelen+1
suchtext	resb maxfilenamelen+1
suchtext2	resb maxfilenamelen+1	;for PICO mode
optbuffer	resb optslen		;buffer for search/replace options and for ^QI
linkbuffersize	equ 4
linkbuffer	resb linkbuffersize

%ifdef LINUX
 giduid		resd 1
 accesstime	resd 1			;0 ie NOW
 changetime	resd 1			;seconds
%else
 uid		resd 1
 gid		resd 1
 accesstime	resd 2
 changetime	resd 2
%endif
;------
perms		resd 1
fstatbuf	resb 64
screenline	resb 256+4*scolorslen	;max possible columns + 4 color ESC seq per line
					;(buffer for displaying a text line)
EmaKiSize	resd 1
EmaKiSrc	resd 1
EmaMark		resd 1
EmaCtrlK	resd 1
EmaCtrlS	resd 1
EmaNumHelper	resd 1
VICmdMode	resd 1
VIbufch		resd 1
VInolinebased	resd 1
PicoSearch	resd 1			;where search started

sedpipeB0	resd 1
sedpipeB1	resd 1
sedpipeC0	resd 1
sedpipeC1	resd 1

mode		resd 1

buffercopysize	equ 1024
buffercopy	resb buffercopysize

level		resd 1			;balance
stackptr	resd 1			;escape recursion
x87		resd 3			;12 byte (need 10 byte for 80 bit BCD)

%ifdef LINUX
max		equ 102400		;valid for NEW created files only
%else
max		equ 10240000		;FIXME: brk in FreeBSD ??
%endif
;-------
text		resb max
sot 		equ (text+1)		;start-of-text

%ifdef CRIPLED_ELF
 bsssize equ $-bssstart
%endif
